"""
Fxitures for tests.
"""
import logging
import warnings
from pathlib import Path
from typing import Any, Callable

import numpy as np
import pandas as pd
import scipy as sp

import lymph
from lymph import diagnose_times
from lymph.modalities import Clinical, Modality, Pathological
from lymph.models import Unilateral
from lymph.types import PatternType

MODALITIES = {
    "CT": Clinical(spec=0.81, sens=0.86),
    "FNA": Pathological(spec=0.95, sens=0.81),
}
RNG = np.random.default_rng(42)


def get_graph(size: str = "large") -> dict[tuple[str, str], list[str]]:
    """Return either a ``"small"``, a ``"medium"`` or a ``"large"`` graph."""
    if size == "small":
        return {
            ("tumor", "T"): ["II", "III"],
            ("lnl", "II"): ["III"],
            ("lnl", "III"): [],
        }

    if size == "medium":
        return {
            ("tumor", "T"): ["I", "II", "III", "IV"],
            ("lnl", "I"): ["II"],
            ("lnl", "II"): ["III"],
            ("lnl", "III"): ["IV"],
            ("lnl", "IV"): [],
        }

    if size == "large":
        return {
            ("tumor", "T"): ["I", "II", "III", "IV", "V", "VII"],
            ("lnl", "I"): [],
            ("lnl", "II"): ["I", "III", "V"],
            ("lnl", "III"): ["IV", "V"],
            ("lnl", "IV"): [],
            ("lnl", "V"): [],
            ("lnl", "VII"): [],
        }

    raise ValueError(f"Unknown graph size: {size}")


def get_logger(
    level: str = logging.INFO,
    handler: logging.Handler = logging.StreamHandler(),
) -> logging.Logger:
    """Return the :py:mod:`lymph` package's logger with a handler."""
    logger = logging.getLogger("lymph")
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger


def _create_random_frozen_dist(
    max_time: int,
    rng: np.random.Generator = RNG,
) -> np.ndarray:
    """Create a random frozen diagnose time distribution."""
    unnormalized = rng.random(size=max_time + 1)
    return unnormalized / np.sum(unnormalized)

def _create_random_parametric_dist(
    max_time: int,
    rng: np.random.Generator = RNG,
) -> diagnose_times.Distribution:
    """Create a binomial diagnose time distribution with random params."""
    def _pmf(support: np.ndarray, p: float = rng.random()) -> np.ndarray:
        return sp.stats.binom.pmf(support, p=p, n=max_time + 1)

    return diagnose_times.Distribution(
        distribution=_pmf,
        max_time=max_time,
    )

def create_random_dist(
    type_: str,
    max_time: int,
    rng: np.random.Generator = RNG,
) -> np.ndarray | Callable:
    """Create a random frozen or parametric distribution."""
    if type_ == "frozen":
        return _create_random_frozen_dist(max_time=max_time, rng=rng)

    if type_ == "parametric":
        return _create_random_parametric_dist(max_time=max_time, rng=rng)

    raise ValueError(f"Unknown distribution type: {type_}")


def create_random_pattern(lnls: list[str]) -> PatternType:
    """Create a random involvement pattern."""
    return {
        lnl: RNG.choice([True, False, None])
        for lnl in lnls
    }


class BinaryUnilateralModelMixin:
    """Mixin class for simple binary model fixture creation."""

    def setUp(self, graph_size: str = "large"):
        """Initialize a simple binary model."""
        self.rng = np.random.default_rng(42)
        self.graph_dict = get_graph(size=graph_size)
        self.model = Unilateral.binary(graph_dict=self.graph_dict)
        self.logger = get_logger(level=logging.INFO)


    def create_random_params(self) -> dict[str, float]:
        """Create random parameters for the model."""
        params = {
            f"{name}_{type_}": self.rng.random()
            for name, edge in self.model.graph.edges.items()
            for type_ in edge.get_params(as_dict=True).keys()
        }
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            params.update({
                f"{t_stage}_{type_}": self.rng.random()
                for t_stage, dist in self.model.get_all_distributions().items()
                for type_ in dist.get_params(as_dict=True).keys()
            })
        return params


    def init_diag_time_dists(self, **dists) -> None:
        """Init the diagnose time distributions."""
        for t_stage, type_ in dists.items():
            self.model.set_distribution(
                t_stage,
                create_random_dist(type_, self.model.max_time, self.rng),
            )


    def load_patient_data(
        self,
        filename: str = "2021-clb-oropharynx.csv",
    ) -> None:
        """Load patient data from a CSV file."""
        filepath = Path(__file__).parent / "data" / filename
        self.raw_data = pd.read_csv(filepath, header=[0,1,2])
        self.model.load_patient_data(self.raw_data, side="ipsi")


class BilateralModelMixin:
    """Mixin for testing the bilateral model."""
    model_kwargs: dict[str, Any] | None = None

    def setUp(self):
        if self.model_kwargs is None:
            self.model_kwargs = {}

        super().setUp()
        self.rng = np.random.default_rng(42)
        self.graph_dict = get_graph("large")
        self.model = lymph.models.Bilateral(graph_dict=self.graph_dict, **self.model_kwargs)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())
        self.logger = get_logger(level=logging.INFO)


    def init_diag_time_dists(self, **dists) -> None:
        """Init the diagnose time distributions."""
        for t_stage, type_ in dists.items():
            self.model.set_distribution(
                t_stage,
                create_random_dist(type_, self.model.max_time, self.rng),
            )


    def create_random_params(self) -> dict[str, float]:
        """Create a random set of parameters."""
        params = self.model.get_params(as_dict=True)

        for name in params:
            params[name] = self.rng.random()

        return params


    def load_patient_data(
        self,
        filename: str = "2021-usz-oropharynx.csv",
    ) -> None:
        """Load patient data from a CSV file."""
        filepath = Path(__file__).parent / "data" / filename
        self.raw_data = pd.read_csv(filepath, header=[0,1,2])
        self.model.load_patient_data(self.raw_data)



class TrinaryFixtureMixin:
    """Mixin class for simple trinary model fixture creation."""

    def setUp(self):
        """Initialize a simple trinary model."""
        self.rng = np.random.default_rng(42)
        self.graph_dict = get_graph(size="large")
        self.model = Unilateral.trinary(graph_dict=self.graph_dict)
        self.logger = get_logger(level=logging.INFO)


    def create_random_params(self) -> dict[str, float]:
        """Create random parameters for the model."""
        params = {
            f"{name}_{type_}": self.rng.random()
            for name, edge in self.model.graph.edges.items()
            for type_ in edge.get_params(as_dict=True).keys()
        }
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            params.update({
                f"{t_stage}_{type_}": self.rng.random()
                for t_stage, dist in self.model.get_all_distributions().items()
                for type_ in dist.get_params(as_dict=True).keys()
            })

        return params


    def init_diag_time_dists(self, **dists) -> None:
        """Init the diagnose time distributions."""
        for t_stage, type_ in dists.items():
            self.model.set_distribution(
                t_stage,
                create_random_dist(type_, self.model.max_time, self.rng),
            )


    def get_modalities_subset(self, names: list[str]) -> dict[str, Modality]:
        """Create a dictionary of modalities."""
        modalities_in_data = {
            "CT": Clinical(spec=0.76, sens=0.81),
            "MRI": Clinical(spec=0.63, sens=0.81),
            "PET": Clinical(spec=0.86, sens=0.79),
            "FNA": Pathological(spec=0.98, sens=0.80),
            "diagnostic_consensus": Clinical(spec=0.86, sens=0.81),
            "pathology": Pathological(spec=1.0, sens=1.0),
            "pCT": Clinical(spec=0.86, sens=0.81),
        }
        return {name: modalities_in_data[name] for name in names}


    def load_patient_data(
        self,
        filename: str = "2021-clb-oropharynx.csv",
    ) -> None:
        """Load patient data from a CSV file."""
        filepath = Path(__file__).parent / "data" / filename
        self.raw_data = pd.read_csv(filepath, header=[0,1,2])
        self.model.load_patient_data(self.raw_data, side="ipsi")



"""
Test the Bayesian Unilateral Model.
"""
import unittest

import numpy as np

from . import fixtures


class BayesianUnilateralModelTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the Bayesian Unilateral Model."""

    def setUp(self):
        super().setUp()
        self.model.set_params(**self.create_random_params())
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def test_state_dist(self):
        """Test the state distribution."""
        bayes_state_dist = self.model.comp_state_dist(mode="BN")
        self.assertTrue(np.isclose(bayes_state_dist.sum(), 1.0))

    def test_obs_dist(self):
        """Test the observation distribution."""
        bayes_obs_dist = self.model.comp_obs_dist(mode="BN")
        self.assertTrue(np.isclose(bayes_obs_dist.sum(), 1.0))

    def test_log_likelihood_smaller_zero(self):
        """Test the likelihood."""
        likelihood = self.model.likelihood(mode="BN")
        self.assertLessEqual(likelihood, 0.)

    def test_likelihood_invalid_params_isinf(self):
        """Make sure the likelihood is `-np.inf` for invalid parameters."""
        random_params = self.create_random_params()
        for name in random_params:
            random_params[name] += 1.
        likelihood = self.model.likelihood(
            given_params=random_params,
            log=True,
            mode="BN",
        )
        self.assertEqual(likelihood, -np.inf)



"""
Module for marginalizing over diagnose times.

The hidden Markov model we implement assumes that every patient started off with a
healthy neck, meaning no lymph node levels harboured any metastases. This is a valid
assumption, but brings with it the issue of determining *how long ago* this likely was.

This module allows the user to define a distribution over abstract time-steps that
indicate for different T-categories how probable a diagnosis at this time-step was.
That allows us to treat T1 and T4 patients fundamentally in the same way, even with the
same parameters, except for the parametrization of their respective distribution over
the time of diagnosis.
"""
from __future__ import annotations

import inspect
import logging
import warnings
from abc import ABC
from typing import Any, Iterable, TypeVar

import numpy as np

from lymph.helper import flatten, popfirst, unflatten_and_split

logger = logging.getLogger(__name__)


class SupportError(Exception):
    """Error that is raised when no support for a distribution is provided."""


class Distribution:
    """Class that provides a way of storing distributions over diagnose times."""
    def __init__(
        self,
        distribution: Iterable[float] | callable,
        max_time: int | None = None,
    ) -> None:
        """Initialize a distribution over diagnose times.

        This object can either be created by passing a parametrized function (e.g.,
        ``scipy.stats`` distribution) or by passing a list of probabilities for each
        diagnose time.

        The signature of the function must be ``func(support, **kwargs)``, where
        ``support`` is the support of the distribution from 0 to ``max_time``. The
        function must return a list of probabilities for each diagnose time.

        Note:
            All arguments except ``support`` must have default values and if some
            parameters have bounds (like the binomial distribution's ``p``), the
            function must raise a ``ValueError`` if the parameter is invalid.

        Since ``max_time`` specifies the support of the distribution (ranging from 0 to
        ``max_time``), it must be provided if a parametrized function is passed. If a
        list of probabilities is passed, ``max_time`` is inferred from the length of the
        list and can be omitted. But an error is raised if the length of the list and
        ``max_time`` + 1 don't match, in case it is accidentally provided.
        """
        if callable(distribution):
            self._init_from_callable(distribution, max_time)
        elif isinstance(distribution, Distribution):
            self._init_from_instance(distribution)
        else:
            self._init_from_frozen(distribution, max_time)


    def _init_from_callable(self, distribution: callable, max_time: int | None = None):
        """Initialize the distribution from a callable distribution."""
        if max_time is None:
            raise ValueError("max_time must be provided if a function is passed")
        if max_time < 0:
            raise ValueError("max_time must be a positive integer")

        self.max_time = max_time
        self._kwargs = self.extract_kwargs(distribution)
        self._func = distribution
        self._frozen = self.pmf


    def _init_from_instance(self, instance: Distribution):
        """Initialize the distribution from another instance."""
        if not instance.is_updateable:
            self._init_from_frozen(instance.pmf, instance.max_time)
        else:
            self.max_time = instance.max_time
            self._kwargs = instance._kwargs.copy()
            self._func = instance._func
            self._frozen = self.pmf


    def _init_from_frozen(self, distribution: Iterable[float], max_time: int | None = None):
        """Initialize the distribution from a frozen distribution."""
        if max_time is None:
            max_time = len(distribution) - 1

        if max_time != len(distribution) - 1:
            raise ValueError(
                f"max_time {max_time} and len of distribution {len(distribution)} "
                "don't match"
            )

        self.max_time = max_time
        self._kwargs = {}
        self._func = None
        self._frozen = self.normalize(distribution)


    @staticmethod
    def extract_kwargs(distribution: callable) -> dict[str, Any]:
        """Extract the keyword arguments from the provided parametric distribution.

        The signature of the provided parametric distribution must be
        ``func(support, **kwargs)``. The first argument is the support of the
        distribution, which is a list or array of integers from 0 to ``max_time``.
        The ``**kwargs`` are keyword parameters that are passed to the function to
        update it.
        """
        kwargs = {}
        # skip the first parameter, which is the support
        skip_first = True
        for name, param in inspect.signature(distribution).parameters.items():
            if skip_first:
                skip_first = False
                continue

            if param.default is inspect.Parameter.empty:
                raise ValueError("All params of the function must be keyword arguments")

            kwargs[name] = param.default

        return kwargs


    def __repr__(self) -> str:
        return f"Distribution({repr(self.pmf.tolist())})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, Distribution):
            return False

        return (
            self.is_updateable == other.is_updateable
            and self._kwargs == other._kwargs
            and np.all(self.pmf == other.pmf)
        )

    def __len__(self) -> int:
        return len(self.support)

    def __hash__(self) -> int:
        """Return a hash of the distribution.

        This is computed from the stored frozen distribution and -- if
        :py:meth:`.is_updateable` returns ``True`` -- the stored keyword arguments of
        the parametric distribution.
        """
        kwarg_tpl = tuple(self._kwargs.items())
        return hash((self.is_updateable, kwarg_tpl, self.pmf.tobytes()))


    @property
    def max_time(self) -> int:
        """Return the maximum time for the distribution."""
        return self.support[-1]

    @max_time.setter
    def max_time(self, value: int) -> None:
        """Set the maximum time for the distribution."""
        if value < 0:
            raise ValueError("max_time must be a positive integer")

        self.support = np.arange(value + 1)
        self._frozen = None


    @staticmethod
    def normalize(distribution: np.ndarray) -> np.ndarray:
        """Normalize a distribution."""
        distribution = np.array(distribution)
        return distribution / np.sum(distribution)


    @property
    def pmf(self) -> np.ndarray:
        """Return the probability mass function of the distribution if it is frozen."""
        if not hasattr(self, "_frozen") or self._frozen is None:
            self._frozen = self.normalize(self._func(self.support, **self._kwargs))
        return self._frozen


    @property
    def is_updateable(self) -> bool:
        """``True`` if instance can be updated via :py:meth:`~set_param`."""
        return self._func is not None


    def get_params(
        self,
        as_dict: bool = True,
        **_kwargs,
    ) -> float | Iterable[float] | dict[str, float]:
        """If updateable, return the dist's ``param`` value or all params in a dict.

        See Also:
            :py:meth:`lymph.diagnose_times.DistributionsUserDict.get_params`
            :py:meth:`lymph.graph.Edge.get_params`
            :py:meth:`lymph.models.Unilateral.get_params`
            :py:meth:`lymph.models.Bilateral.get_params`
        """
        if not self.is_updateable:
            warnings.warn("Distribution is not updateable, returning empty dict")
            return {} if as_dict else None

        return self._kwargs if as_dict else self._kwargs.values()


    def set_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Update distribution by setting its parameters and storing the frozen PMF.

        Parameters can be set via positional arguments - which are used up one by one
        in the order they are provided and are then returned - or keyword arguments.
        Keyword arguments override positional arguments. If the distribution is not
        updateable, a warning is issued and all args and kwargs are returned.

        If any of the parameters is invalid, a ``ValueError`` is raised and the original
        parameters are restored.
        """
        if not self.is_updateable:
            warnings.warn("Distribution is not updateable, ignoring parameters")
            return args

        old_kwargs = self._kwargs.copy()

        for name, value in self._kwargs.items():
            first, args = popfirst(args)
            self._kwargs[name] = first or kwargs.get(name, value)
            if hasattr(self, "_frozen"):
                del self._frozen

        try:
            _ = self.pmf
        except ValueError as val_err:
            self._kwargs = old_kwargs
            raise ValueError("Invalid params provided to distribution") from val_err

        return args


    def draw_diag_times(
        self,
        num: int | None = None,
        rng: np.random.Generator | None = None,
        seed: int = 42,
    ) -> np.ndarray:
        """Draw ``num`` samples of diagnose times from the stored PMF.

        A random number generator can be provided as ``rng``. If ``None``, a new one
        is initialized with the given ``seed`` (or ``42``, by default).
        """
        if rng is None:
            rng = np.random.default_rng(seed)

        return rng.choice(a=self.support, p=self.pmf, size=num)



DC = TypeVar("DC", bound="Composite")

class Composite(ABC):
    """Abstract base class implementing the composite pattern for distributions.

    Any class inheriting from this class should be able to handle the definition of
    distributions over diagnosis times.

    >>> class MyComposite(Composite):
    ...     pass
    >>> leaf1 = MyComposite(is_distribution_leaf=True, max_time=1)
    >>> leaf2 = MyComposite(is_distribution_leaf=True, max_time=1)
    >>> leaf3 = MyComposite(is_distribution_leaf=True, max_time=1)
    >>> branch1 = MyComposite(distribution_children={"L1": leaf1, "L2": leaf2})
    >>> branch2 = MyComposite(distribution_children={"L3": leaf3})
    >>> root = MyComposite(distribution_children={"B1": branch1, "B2": branch2})
    >>> root.set_distribution("T1", Distribution([0.1, 0.9]))
    >>> root.get_distribution("T1")
    Distribution([0.1, 0.9])
    >>> leaf1.get_distribution("T1")
    Distribution([0.1, 0.9])
    """
    _max_time: int
    _distributions: dict[str, Distribution]    # only for leaf nodes
    _distribution_children: dict[str, Composite]

    def __init__(
        self: DC,
        max_time: int | None = None,
        distribution_children: dict[str, Composite] | None = None,
        is_distribution_leaf: bool = False,
    ) -> None:
        """Initialize the distribution composite."""
        if distribution_children is None:
            distribution_children = {}

        if is_distribution_leaf:
            self._distributions = {}
            self._distribution_children = {}    # ignore any provided children
            self.max_time = max_time            # only set max_time in leaf

        self._distribution_children = distribution_children

        super().__init__()


    @property
    def _is_distribution_leaf(self: DC) -> bool:
        """Return whether the object is a leaf node w.r.t. distributions."""
        if len(self._distribution_children) > 0:
            return False

        if not hasattr(self, "_distributions"):
            raise AttributeError(f"{self} has no children and no distributions.")

        return True


    @property
    def max_time(self: DC) -> int:
        """Return the maximum time for the distributions."""
        if self._is_distribution_leaf:
            are_all_equal = True
            for dist in self._distributions.values():
                are_equal = dist.max_time == self._max_time
                if not are_equal:
                    dist.max_time = self._max_time
                are_all_equal &= are_equal

            if not are_all_equal:
                warnings.warn(f"Not all max_times were equal. Set all to {self._max_time}")

            return self._max_time

        max_times = [child.max_time for child in self._distribution_children.values()]
        if len(set(max_times)) > 1:
            warnings.warn("Not all max_times are equal. Returning the first one.")

        return max_times[0]

    @max_time.setter
    def max_time(self: DC, value: int) -> None:
        """Set the maximum time for the distributions."""
        if self._is_distribution_leaf:
            if value is None:
                raise ValueError("max_time must be provided if the composite is a leaf")

            if value < 0:
                raise ValueError("max_time must be a positive integer")

            self._max_time = value
            for dist in self._distributions.values():
                dist.max_time = value

        else:
            for child in self._distribution_children.values():
                child.max_time = value


    @property
    def t_stages(self: DC) -> list[str]:
        """Return the T-stages for which distributions are defined."""
        return list(self.get_all_distributions().keys())


    def get_distribution(self: DC, t_stage: str) -> Distribution:
        """Return the distribution for the given ``t_stage``."""
        return self.get_all_distributions()[t_stage]


    def get_all_distributions(self: DC) -> dict[str, Distribution]:
        """Return all distributions.

        This will issue a warning if it finds that not all distributions of the
        composite are equal. Note that it will always return the distributions of the
        first child. This means one should NOT try to set the distributions via the
        returned dictionary of this method. Instead, use the
        :py:meth:`.set_distribution` method.
        """
        if self._is_distribution_leaf:
            return self._distributions

        child_keys = list(self._distribution_children.keys())
        first_child = self._distribution_children[child_keys[0]]
        first_distributions = first_child.get_all_distributions()
        are_all_equal = True
        for key in child_keys[1:]:
            other_child = self._distribution_children[key]
            are_all_equal &= first_distributions == other_child.get_all_distributions()

        if not are_all_equal:
            warnings.warn("Not all distributions are equal. Returning the first one.")

        return first_distributions


    def set_distribution(
        self: DC,
        t_stage: str,
        distribution: Distribution | Iterable[float] | callable,
    ) -> None:
        """Set/update the distribution for the given ``t_stage``."""
        if self._is_distribution_leaf:
            self._distributions[t_stage] = Distribution(distribution, self.max_time)

        else:
            for child in self._distribution_children.values():
                child.set_distribution(t_stage, distribution)


    def del_distribution(self: DC, t_stage: str) -> None:
        """Delete the distribution for the given ``t_stage``."""
        if self._is_distribution_leaf:
            del self._distributions[t_stage]

        else:
            for child in self._distribution_children.values():
                child.del_distribution(t_stage)


    def replace_all_distributions(self: DC, distributions: dict[str, Distribution]) -> None:
        """Replace all distributions with the given ones."""
        if self._is_distribution_leaf:
            self._distributions = {}
            for t_stage, distribution in distributions.items():
                self.set_distribution(t_stage, distribution)

        else:
            for child in self._distribution_children.values():
                child.replace_all_distributions(distributions)


    def clear_distributions(self: DC) -> None:
        """Remove all distributions."""
        if self._is_distribution_leaf:
            self._distributions.clear()

        else:
            for child in self._distribution_children.values():
                child.clear_distributions()


    def distributions_hash(self: DC) -> int:
        """Return a hash of all distributions."""
        hash_res = 0
        if self._is_distribution_leaf:
            for t_stage, distribution in self._distributions.items():
                hash_res = hash((hash_res, t_stage, hash(distribution)))

        else:
            for child in self._distribution_children.values():
                hash_res = hash((hash_res, child.distributions_hash()))

        return hash_res


    def get_distribution_params(
        self: DC,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of all distributions."""
        params = {}

        if self._is_distribution_leaf:
            for t_stage, distribution in self._distributions.items():
                if not distribution.is_updateable:
                    continue
                params[t_stage] = distribution.get_params(as_flat=as_flat)
        else:
            child_keys = list(self._distribution_children.keys())
            first_child = self._distribution_children[child_keys[0]]
            params = first_child.get_distribution_params(as_flat=as_flat)
            are_all_equal = True
            for key in child_keys[1:]:
                other_child = self._distribution_children[key]
                other_params = other_child.get_distribution_params(as_flat=as_flat)
                are_all_equal &= params == other_params

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def set_distribution_params(self: DC, *args: float, **kwargs: float) -> tuple[float]:
        """Set the parameters of all distributions."""
        if self._is_distribution_leaf:
            kwargs, global_kwargs = unflatten_and_split(
                kwargs, expected_keys=self._distributions.keys()
            )
            for t_stage, distribution in self._distributions.items():
                if not distribution.is_updateable:
                    continue
                t_stage_kwargs = global_kwargs.copy()
                t_stage_kwargs.update(kwargs.get(t_stage, {}))
                args = distribution.set_params(*args, **t_stage_kwargs)
            # in leafs, use up args one by one
            return args

        kwargs, global_kwargs = unflatten_and_split(
            kwargs, expected_keys=self._distribution_children.keys()
        )
        for key, child in self._distribution_children.items():
            child_kwargs = global_kwargs.copy()
            child_kwargs.update(kwargs.get(key, {}))
            rem_args = child.set_distribution_params(*args, **child_kwargs)
        # in branches, distribute all args to children
        return rem_args



"""
Type aliases and protocols used in the lymph package.
"""
from abc import ABC, abstractmethod
from typing import Iterable, Literal, Protocol, TypeVar

import numpy as np
import pandas as pd
from pandas._libs.missing import NAType


class HasSetParams(Protocol):
    """Protocol for classes that have a ``set_params`` method."""
    def set_params(self, *args: float, **kwargs: float) -> tuple[float]:
        ...


class HasGetParams(Protocol):
    """Protocol for classes that have a ``get_params`` method."""
    def get_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> tuple[float] | dict[str, float]:
        ...


PatternType = dict[str, bool | NAType | None]
"""Type alias for an involvement pattern.

An involvement pattern is a dictionary with keys for the lymph node levels and values
for the involvement of the respective lymph nodes. The values are either True, False,
or None, which means that the involvement is unknown.

>>> pattern = {"I": True, "II": False, "III": None}
"""

DiagnoseType = dict[str, PatternType]
"""Type alias for a diagnose, which is an involvement pattern per diagnostic modality.

>>> diagnose = {
...     "CT": {"I": True, "II": False, "III": None},
...     "MRI": {"I": True, "II": True, "III": None},
... }
"""


M = TypeVar("M", bound="Model")

class Model(ABC):
    """Abstract base class for models.

    This class provides a scaffold for the methods that any model for lymphatic
    tumor progression should implement.
    """
    @abstractmethod
    def get_params(
        self: M,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the model.

        The parameters are returned as a dictionary if ``as_dict`` is True, and as
        an iterable of floats otherwise. The argument ``as_flat`` determines whether
        the returned dict is flat or nested. This is helpful, because a model may call
        the ``get_params`` method of other instances, which can be fused to get a
        flat dictionary.
        """
        raise NotImplementedError

    def get_num_dims(self: M, mode: Literal["HMM", "BN"] = "HMM") -> int:
        """Return the number of dimensions of the parameter space.

        A hidden Markov model (``mode="HMM"``) typically has more parameters than a
        Bayesian network (``mode="BN"``), because it we need parameters for the
        distributions over diagnosis times. Your can read more about that in the
        :py:mod:`lymph.diagnose_times` module.
        """
        # pylint: disable=no-member
        num = len(self.get_params())
        if mode == "BN":
            num -= len(self.get_distribution_params())
        return num

    @abstractmethod
    def set_params(self: M, *args: float, **kwargs: float) -> tuple[float]:
        """Set the parameters of the model.

        The parameters may be passed as positional or keyword arguments. The positional
        arguments are used up one by one by the ``set_params`` methods the model calls.
        Keyword arguments override the positional arguments.
        """
        raise NotImplementedError

    @abstractmethod
    def load_patient_data(
        self: M,
        patient_data: pd.DataFrame,
    ) -> None:
        """Load patient data in `LyProX`_ format into the model.

        .. _LyProX: https://lyprox.org/
        """
        raise NotImplementedError

    @abstractmethod
    def likelihood(
        self: M,
        given_params: Iterable[float] | dict[str, float] | None = None,
        log: bool = True,
    ) -> float:
        """Return the likelihood of the model given the parameters.

        The likelihood is returned in log space if ``log`` is True, and in linear space
        otherwise. The parameters may be passed as positional or keyword arguments.
        They are then passed to the :py:meth:`set_params` method first.
        """
        raise NotImplementedError

    @abstractmethod
    def risk(
        self,
        involvement: PatternType | None = None,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: dict[str, PatternType] | None = None,
    ) -> float | np.ndarray:
        """Return the risk of ``involvement``, given the parameters and diagnoses."""
        raise NotImplementedError



"""
This module implements the core classes to model lymphatic tumor progression.
"""
from lymph.models.bilateral import Bilateral
from lymph.models.midline import Midline
from lymph.models.unilateral import Unilateral

__all__ = ["Unilateral", "Bilateral", "Midline"]



"""
Make doctests in the lymph package discoverable by unittest.
"""
import doctest
import unittest

from lymph import diagnose_times, graph, helper, matrix, modalities
from lymph.models import bilateral, unilateral


def load_tests(loader, tests: unittest.TestSuite, ignore):
    """Load doctests from the lymph package."""
    tests.addTests(doctest.DocTestSuite(diagnose_times))
    tests.addTests(doctest.DocTestSuite(graph))
    tests.addTests(doctest.DocTestSuite(helper))
    tests.addTests(doctest.DocTestSuite(matrix))
    tests.addTests(doctest.DocTestSuite(modalities))

    tests.addTests(doctest.DocTestSuite(unilateral))
    tests.addTests(doctest.DocTestSuite(bilateral))
    return tests



"""Unit tests for the Node classes."""
import unittest

from lymph import graph


class BinaryLymphNodeLevelTestCase(unittest.TestCase):
    """Test case for a binary lymph node level."""

    def setUp(self) -> None:
        super().setUp()
        self.initial_name = "test"
        self.initial_state = 0
        self.lnl = graph.LymphNodeLevel.binary(
            self.initial_name,
            state=self.initial_state,
        )

    def test_binary_init(self) -> None:
        """Test the binary node initialization."""
        self.assertEqual(self.lnl.name, self.initial_name)
        self.assertEqual(self.lnl.state, self.initial_state)
        self.assertEqual(self.lnl.allowed_states, [0, 1])
        self.assertTrue(self.lnl.is_binary)

    def test_str(self) -> None:
        """Test the string representation of the node."""
        self.assertEqual(str(self.lnl), f"binary LNL '{self.initial_name}'")

    def test_repr(self) -> None:
        """Test if the node can be recreated from repr."""
        # pylint: disable=eval-used
        recreated_lnl = eval(repr(self.lnl), {"LymphNodeLevel": graph.LymphNodeLevel})
        self.assertEqual(self.lnl.name, recreated_lnl.name)
        self.assertEqual(self.lnl.state, recreated_lnl.state)
        self.assertEqual(self.lnl.allowed_states, recreated_lnl.allowed_states)



"""
Test the midline model for the binary case.
"""
import unittest
from typing import Literal

import numpy as np

from lymph import models

from . import fixtures


class MidlineSetParamsTestCase(unittest.TestCase):
    """Check that the complex parameter assignment works correctly."""

    def setUp(
        self,
        seed: int = 42,
        graph_size: Literal["small", "medium", "large"] = "small",
        use_mixing: bool = True,
        use_central: bool = True,
        is_symmetric: dict[str, bool] | None = None,
    ) -> None:
        super().setUp()
        self.rng = np.random.default_rng(seed)
        graph_dict = fixtures.get_graph(graph_size)
        if is_symmetric is None:
            is_symmetric = {"tumor_spread": False, "lnl_spread": True}

        self.model = models.Midline.trinary(
            graph_dict=graph_dict,
            is_symmetric=is_symmetric,
            use_mixing=use_mixing,
            use_central=use_central,
            use_midext_evo=False,
        )


    def test_init(self) -> None:
        """Check some basic attributes."""
        self.assertTrue(self.model.use_central)
        self.assertTrue(self.model.use_mixing)
        self.assertTrue(self.model.is_trinary)


    def test_set_spread_params(self) -> None:
        """Check that the complex parameter assignment works correctly."""
        params_to_set = {k: self.rng.uniform() for k in self.model.get_params().keys()}
        self.model.set_params(**params_to_set)

        self.assertEqual(
            self.model.central.ipsi.get_tumor_spread_params(),
            self.model.central.contra.get_tumor_spread_params(),
        )
        self.assertEqual(
            self.model.central.ipsi.get_lnl_spread_params(),
            self.model.central.contra.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.central.contra.get_lnl_spread_params(),
            self.model.ext.ipsi.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.ext.ipsi.get_lnl_spread_params(),
            self.model.noext.ipsi.get_lnl_spread_params(),
        )
        self.assertEqual(
            self.model.ext.ipsi.get_tumor_spread_params(),
            self.model.noext.ipsi.get_tumor_spread_params(),
        )



from __future__ import annotations

import logging
import warnings
from typing import Any, Iterable, Literal

import numpy as np
import pandas as pd

from lymph import diagnose_times, matrix, modalities, models, types
from lymph.helper import (
    add_or_mult,
    draw_diagnoses,
    early_late_mapping,
    flatten,
    popfirst,
    unflatten_and_split,
)
from lymph.types import DiagnoseType, PatternType

warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)
logger = logging.getLogger(__name__)


EXT_COL = ("tumor", "1", "extension")
CENTRAL_COL = ("tumor", "1", "central")



class Midline(
    diagnose_times.Composite,
    modalities.Composite,
    types.Model,
):
    """Models metastatic progression bilaterally with tumor lateralization.

    Model a bilateral lymphatic system where an additional risk factor can
    be provided in the data: Whether or not the primary tumor extended over the
    mid-sagittal line, or is located on the mid-saggital line.

    It is reasonable to assume (and supported by data) that an extension of the primary
    tumor significantly increases the risk for metastatic spread to the contralateral
    side of the neck. This class attempts to capture this using a simple
    assumption: We assume that the probability of spread to the contralateral
    side for patients *with* midline extension is larger than for patients
    *without* it, but smaller than the probability of spread to the ipsilateral
    side. Formally:

    .. math::
        b_c^{\\in} = \\alpha \\cdot b_i + (1 - \\alpha) \\cdot b_c^{\\not\\in}

    where :math:`b_c^{\\in}` is the probability of spread from the primary tumor
    to the contralateral side for patients with midline extension, and
    :math:`b_c^{\\not\\in}` for patients without. :math:`\\alpha` is the linear
    mixing parameter.
    """
    def __init__(
        self,
        graph_dict: dict[tuple[str], list[str]],
        is_symmetric: dict[str, bool] | None = None,
        use_mixing: bool = True,
        use_central: bool = True,
        use_midext_evo: bool = True,
        marginalize_unknown: bool = True,
        unilateral_kwargs: dict[str, Any] | None = None,
        **_kwargs
    ):
        """Initialize the model.

        The class is constructed in a similar fashion to the :py:class:`~.Bilateral`:
        That class contains one :py:class:`~.Unilateral` for each side of the neck,
        while this class will contain several instances of :py:class:`~.Bilateral`,
        one for the ipsilateral side and two to three for the the contralateral side
        covering the cases a) no midline extension, b) midline extension, and c)
        central tumor location.

        Added keyword arguments in this constructor are ``use_mixing``, which controls
        whether to use the above described mixture of spread parameters from tumor to
        the LNLs. And ``use_central``, which controls whether to use a third
        :py:class:`~.Bilateral` model for the case of a central tumor location.

        The parameter ``use_midext_evo`` decides whether the tumor's midline extions
        should be considered a random variable, in which case it is evolved like the
        state of the LNLs, or not.

        With ``marginalize_unknown`` (default: ``True``), the model will also load
        patients with unknown midline extension status into the model and marginalize
        over their state of midline extension when computing the likelihood. This extra
        data is stored in a :py:class:`~.Bilateral` instance accessible via the
        attribute ``"unknown"``. Note that this bilateral instance does not get updated
        parameters or any other kind of attention. It is solely used to store the data
        and generate diagnose matrices for those data.

        The ``unilateral_kwargs`` are passed to all bilateral models.

        See Also:
            :py:class:`Bilateral`: Two to four of these are held as attributes by this
            class. One for the case of a mid-sagittal extension of the primary
            tumor, one for the case of no such extension, (possibly) one for the case of
            a central/symmetric tumor, and (possibly) one for the case of unknown
            midline extension status.
        """
        if is_symmetric is None:
            is_symmetric = {}

        is_symmetric["tumor_spread"] = is_symmetric.get("tumor_spread", False)
        is_symmetric["lnl_spread"] = is_symmetric.get("lnl_spread", True)

        if is_symmetric["tumor_spread"]:
            raise ValueError(
                "If you want the tumor spread to be symmetric, consider using the "
                "Bilateral class."
            )
        self.is_symmetric = is_symmetric

        self.ext = models.Bilateral(
            graph_dict=graph_dict,
            unilateral_kwargs=unilateral_kwargs,
            is_symmetric=self.is_symmetric,
        )
        self.noext = models.Bilateral(
            graph_dict=graph_dict,
            unilateral_kwargs=unilateral_kwargs,
            is_symmetric=self.is_symmetric,
        )

        self.use_midext_evo = use_midext_evo
        if self.use_midext_evo and use_central:
            raise ValueError(
                "Evolution to central tumor not yet implemented. Choose to use either "
                "the central model or the midline extension evolution."
                # Actually, this shouldn't be too hard, but we still need to think
                # about it for a bit.
            )
        other_children = {}
        if use_central:
            self._central = models.Bilateral(
                graph_dict=graph_dict,
                unilateral_kwargs=unilateral_kwargs,
                is_symmetric={
                    "tumor_spread": True,
                    "lnl_spread": self.is_symmetric["lnl_spread"],
                },
            )
            other_children["central"] = self.central

        if marginalize_unknown:
            self._unknown = models.Bilateral(
                graph_dict=graph_dict,
                unilateral_kwargs=unilateral_kwargs,
                is_symmetric=self.is_symmetric,
            )
            other_children["unknown"] = self._unknown

        if use_mixing:
            self.mixing_param = 0.

        self.midext_prob = 0.

        diagnose_times.Composite.__init__(
            self,
            distribution_children={"ext": self.ext, "noext": self.noext, **other_children},
            is_distribution_leaf=False,
        )
        modalities.Composite.__init__(
            self,
            modality_children={"ext": self.ext, "noext": self.noext, **other_children},
            is_modality_leaf=False,
        )


    @classmethod
    def trinary(cls, *args, **kwargs) -> Midline:
        """Create a trinary model."""
        unilateral_kwargs = kwargs.pop("unilateral_kwargs", {})
        unilateral_kwargs["allowed_states"] = [0, 1, 2]
        return cls(*args, unilateral_kwargs=unilateral_kwargs, **kwargs)


    @property
    def is_trinary(self) -> bool:
        """Return whether the model is trinary."""
        if self.ext.is_trinary != self.noext.is_trinary:
            raise ValueError("The bilateral models must have the same trinary status.")

        if self.use_central and self.central.is_trinary != self.ext.is_trinary:
            raise ValueError("The bilateral models must have the same trinary status.")

        return self.ext.is_trinary


    @property
    def midext_prob(self) -> float:
        """Return the probability of midline extension."""
        if hasattr(self, "_midext_prob"):
            return self._midext_prob
        return 0.

    @midext_prob.setter
    def midext_prob(self, value: float) -> None:
        """Set the probability of midline extension."""
        if value is not None and not 0. <= value <= 1.:
            raise ValueError("The midline extension prob must be in the range [0, 1].")
        self._midext_prob = value


    @property
    def mixing_param(self) -> float | None:
        """Return the mixing parameter."""
        if hasattr(self, "_mixing_param"):
            return self._mixing_param
        return None

    @mixing_param.setter
    def mixing_param(self, value: float) -> None:
        """Set the mixing parameter."""
        if value is not None and not 0. <= value <= 1.:
            raise ValueError("The mixing parameter must be in the range [0, 1].")
        self._mixing_param = value

    @property
    def use_mixing(self) -> bool:
        """Return whether the model uses a mixing parameter."""
        return hasattr(self, "_mixing_param")

    @property
    def use_central(self) -> bool:
        """Return whether the model uses a central model."""
        return hasattr(self, "_central")

    @property
    def central(self) -> models.Bilateral:
        """Return the central model."""
        if self.use_central:
            return self._central
        raise AttributeError("This instance does not account for central tumors.")

    @property
    def marginalize_unknown(self) -> bool:
        """Return whether the model marginalizes over unknown midline extension."""
        return hasattr(self, "_unknown")

    @property
    def unknown(self) -> models.Bilateral:
        """Return the model storing the patients with unknown midline extension."""
        if self.marginalize_unknown:
            return self._unknown
        raise AttributeError(
            "This instance does not marginalize over unknown midline extension."
        )


    def get_tumor_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> dict[str, float] | Iterable[float]:
        """Return the tumor spread parameters of the model.

        If the model uses the mixing parameter, the returned params will contain the
        ipsilateral spread from tumor to LNLs, the contralateral ones for the case of
        no midline extension, and the mixing parameter. Otherwise, it will contain the
        contralateral params for the cases of present and absent midline extension.
        """
        params = {}
        params["ipsi"] = self.ext.ipsi.get_tumor_spread_params(as_flat=as_flat)

        if self.use_mixing:
            params["contra"] = self.noext.contra.get_tumor_spread_params(as_flat=as_flat)
            params["mixing"] = self.mixing_param
        else:
            params["noext"] = {
                "contra": self.noext.contra.get_tumor_spread_params(as_flat=as_flat)
            }
            params["ext"] = {
                "contra": self.ext.contra.get_tumor_spread_params(as_flat=as_flat)
            }

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def get_lnl_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> dict[str, float] | Iterable[float]:
        """Return the LNL spread parameters of the model.

        Depending on the value of ``is_symmetric["lnl_spread"]``, the returned params
        may contain only one set of spread parameters (if ``True``) or one for the ipsi-
        and one for the contralateral side (if ``False``).
        """
        ext_lnl_params = self.ext.get_lnl_spread_params(as_flat=False)
        noext_lnl_params = self.noext.get_lnl_spread_params(as_flat=False)

        if ext_lnl_params != noext_lnl_params:
            raise ValueError(
                "LNL spread params not synched between ext and noext models. "
                "Returning the ext params."
            )

        if self.use_central:
            central_lnl_params = self.central.get_lnl_spread_params(as_flat=False)
            if central_lnl_params != ext_lnl_params:
                warnings.warn(
                    "LNL spread params not synched between central and ext models. "
                    "Returning the ext params."
                )

        if as_flat or not as_dict:
            ext_lnl_params = flatten(ext_lnl_params)

        return ext_lnl_params if as_dict else ext_lnl_params.values()


    def get_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> dict[str, float] | Iterable[float]:
        """Return the spread parameters of the model.

        This combines the returned values from the calls to
        :py:meth:`get_tumor_spread_params` and :py:meth:`get_lnl_spread_params`.
        """
        params = self.get_tumor_spread_params(as_flat=False)
        lnl_spread_params = self.get_lnl_spread_params(as_flat=False)

        if self.is_symmetric["lnl_spread"]:
            params.update(lnl_spread_params)
        else:
            if "contra" not in params:
                params["contra"] = {}
            params["ipsi"].update(lnl_spread_params["ipsi"])
            params["contra"].update(lnl_spread_params["contra"])

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def get_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return all the parameters of the model.

        This includes the spread parameters from the call to :py:meth:`get_spread_params`
        and the distribution parameters from the call to :py:meth:`get_distribution_params`.
        """
        params = self.get_spread_params(as_flat=as_flat)
        params["mixing"] = self.mixing_param
        params["midext_prob"] = self.midext_prob
        params.update(self.get_distribution_params(as_flat=as_flat))

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def set_tumor_spread_params(
        self, *args: float, **kwargs: float,
    ) -> Iterable[float] | dict[str, float]:
        """Set the spread parameters of the midline model.

        In analogy to the :py:meth:`get_tumor_spread_params` method, this method sets
        the parameters describing how the tumor spreads to the LNLs. How many params
        to provide to this model depends on the value of the ``use_mixing`` and the
        ``use_central`` attributes. Have a look at what the
        :py:meth:`get_tumor_spread_params` method returns for an insight in what you
        can provide.
        """
        kwargs, global_kwargs = unflatten_and_split(
            kwargs, expected_keys=["ipsi", "noext", "ext", "contra"],
        )

        # first, take care of ipsilateral tumor spread (same for all models)
        ipsi_kwargs = global_kwargs.copy()
        ipsi_kwargs.update(kwargs.get("ipsi", {}))
        if self.use_central:
            self.central.set_spread_params(*args, **ipsi_kwargs)
        self.ext.ipsi.set_tumor_spread_params(*args, **ipsi_kwargs)
        args = self.noext.ipsi.set_tumor_spread_params(*args, **ipsi_kwargs)

        # then, take care of contralateral tumor spread
        if self.use_mixing:
            contra_kwargs = global_kwargs.copy()
            contra_kwargs.update(kwargs.get("contra", {}))
            args = self.noext.contra.set_tumor_spread_params(*args, **contra_kwargs)
            mixing_param, args = popfirst(args)
            mixing_param = global_kwargs.get("mixing", mixing_param) or self.mixing_param
            self.mixing_param = global_kwargs.get("mixing", mixing_param)

            ext_contra_kwargs = {}
            for (key, ipsi_param), noext_contra_param in zip(
                self.ext.ipsi.get_tumor_spread_params().items(),
                self.noext.contra.get_tumor_spread_params().values(),
            ):
                ext_contra_kwargs[key] = (
                    self.mixing_param * ipsi_param
                    + (1. - self.mixing_param) * noext_contra_param
                )
            self.ext.contra.set_tumor_spread_params(**ext_contra_kwargs)

        else:
            noext_contra_kwargs = global_kwargs.copy()
            noext_contra_kwargs.update(kwargs.get("noext", {}).get("contra", {}))
            args = self.noext.contra.set_tumor_spread_params(*args, **noext_contra_kwargs)

            ext_contra_kwargs = global_kwargs.copy()
            ext_contra_kwargs.update(kwargs.get("ext", {}).get("contra", {}))
            args = self.ext.contra.set_tumor_spread_params(*args, **ext_contra_kwargs)

        return args


    def set_lnl_spread_params(self, *args: float, **kwargs: float) -> Iterable[float]:
        """Set the LNL spread parameters of the midline model.

        This works exactly like the :py:meth:`.Bilateral.set_lnl_spread_params` for the
        user, but under the hood, the parameters also need to be distributed to two or
        three instances of :py:class:`~.Bilateral` depending on the value of the
        ``use_central`` attribute.
        """
        kwargs, global_kwargs = unflatten_and_split(
            kwargs, expected_keys=["ipsi", "noext", "ext", "contra"],
        )
        ipsi_kwargs = global_kwargs.copy()
        ipsi_kwargs.update(kwargs.get("ipsi", {}))

        if self.is_symmetric["lnl_spread"]:
            if self.use_central:
                self.central.ipsi.set_lnl_spread_params(*args, **global_kwargs)
                self.central.contra.set_lnl_spread_params(*args, **global_kwargs)
            self.ext.ipsi.set_lnl_spread_params(*args, **global_kwargs)
            self.ext.contra.set_lnl_spread_params(*args, **global_kwargs)
            self.noext.ipsi.set_lnl_spread_params(*args, **global_kwargs)
            args = self.noext.contra.set_lnl_spread_params(*args, **global_kwargs)

        else:
            if self.use_central:
                self.central.ipsi.set_lnl_spread_params(*args, **ipsi_kwargs)
            self.ext.ipsi.set_lnl_spread_params(*args, **ipsi_kwargs)
            args = self.noext.ipsi.set_lnl_spread_params(*args, **ipsi_kwargs)

            contra_kwargs = global_kwargs.copy()
            contra_kwargs.update(kwargs.get("contra", {}))
            if self.use_central:
                self.central.contra.set_lnl_spread_params(*args, **contra_kwargs)
            self.ext.contra.set_lnl_spread_params(*args, **contra_kwargs)
            args = self.noext.contra.set_lnl_spread_params(*args, **contra_kwargs)

        return args


    def set_spread_params(self, *args: float, **kwargs: float) -> Iterable[float]:
        """Set the spread parameters of the midline model."""
        args = self.set_tumor_spread_params(*args, **kwargs)
        return self.set_lnl_spread_params(*args, **kwargs)


    def set_params(
        self, *args: float, **kwargs: float,
    ) -> Iterable[float] | dict[str, float]:
        """Set all parameters of the model.

        Combines the calls to :py:meth:`.set_spread_params` and
        :py:meth:`.set_distribution_params`.
        """
        args = self.set_spread_params(*args, **kwargs)
        first, args = popfirst(args)
        self.midext_prob = kwargs.get("midext_prob", first) or self.midext_prob
        return self.set_distribution_params(*args, **kwargs)


    def load_patient_data(
        self,
        patient_data: pd.DataFrame,
        mapping: callable = early_late_mapping,
    ) -> None:
        """Load patient data into the model.

        This amounts to sorting the patients into three bins:

        1. Patients whose tumor is clearly laterlaized, meaning the column
           ``("tumor", "1", "extension")`` reports ``False``. These get assigned to
           the :py:attr:`.noext` attribute.
        2. Those with a central tumor, indicated by ``True`` in the column
           ``("tumor", "1", "central")``. If the :py:attr:`.use_central` attribute is
           set to ``True``, these patients are assigned to the :py:attr:`.central`
           model. Otherwise, they are assigned to the :py:attr:`.ext` model.
        3. The rest, which amounts to patients whose tumor extends over the mid-sagittal
           line but is not central, i.e., symmetric w.r.t to the mid-sagittal line.
           These are assigned to the :py:attr:`.ext` model.

        The split data is sent to the :py:meth:`.Bilateral.load_patient_data` method of
        the respective models.
        """
        # pylint: disable=singleton-comparison
        is_lateralized = patient_data[EXT_COL] == False
        has_extension = patient_data[EXT_COL] == True
        is_unknown = patient_data[EXT_COL].isna()
        self.noext.load_patient_data(patient_data[is_lateralized], mapping)

        if self.use_central:
            is_central = patient_data[CENTRAL_COL] == True
            has_extension = has_extension & ~is_central
            self.central.load_patient_data(patient_data[is_central], mapping)

        self.ext.load_patient_data(patient_data[has_extension], mapping)

        if self.marginalize_unknown and is_unknown.sum() > 0:
            self.unknown.load_patient_data(patient_data[is_unknown], mapping)
        else:
            warnings.warn(
                f"Discarding {is_unknown.sum()} patients where midline extension "
                "is unknown."
            )


    def comp_midext_evolution(self) -> np.ndarray:
        """Evolve only the state of the midline extension."""
        midext_states = np.zeros(shape=(self.max_time + 1, 2), dtype=float)
        midext_states[0,0] = 1.

        midextransition_matrix = np.array([
            [1 - self.midext_prob, self.midext_prob],
            [0.                  , 1.              ],
        ])

        # compute involvement for all time steps
        for i in range(len(midext_states)-1):
            midext_states[i+1,:] = midext_states[i,:] @ midextransition_matrix
        return midext_states


    def comp_contra_dist_evolution(self) -> tuple[np.ndarray, np.ndarray]:
        """Evolve contra side as mixture of with & without midline extension."""
        noext_contra_dist_evo = np.zeros(
            shape=(self.max_time + 1, len(self.noext.contra.graph.state_list))
        )
        noext_contra_dist_evo[0,0] = 1.

        ext_contra_dist_evo = np.zeros(
            shape=(self.max_time + 1, len(self.ext.contra.graph.state_list))
        )
        if not self.use_midext_evo:
            noext_contra_dist_evo[0,0] = (1. - self.midext_prob)
            ext_contra_dist_evo[0,0] = self.midext_prob

        for t in range(self.max_time):
            # When evolving over the midline extension state, there's a chance at any
            # time step that the tumor grows over the midline and starts spreading to
            # the contralateral side more aggressively.
            if self.use_midext_evo:
                noext_contra_dist_evo[t+1] = (
                    (1. - self.midext_prob) * noext_contra_dist_evo[t]
                ) @ self.noext.contra.transition_matrix()
                ext_contra_dist_evo[t+1] = (
                    self.midext_prob * noext_contra_dist_evo[t]
                    + ext_contra_dist_evo[t]
                ) @ self.ext.contra.transition_matrix()

            # When we do not evolve, the tumor is considered lateralized or extending
            # over the midline from the start.
            else:
                noext_contra_dist_evo[t+1] = (
                    noext_contra_dist_evo[t] @ self.noext.contra.transition_matrix()
                )
                ext_contra_dist_evo[t+1] = (
                    ext_contra_dist_evo[t] @ self.ext.contra.transition_matrix()
                )

        return noext_contra_dist_evo, ext_contra_dist_evo


    def _hmm_likelihood(self, log: bool = True, for_t_stage: str | None = None) -> float:
        """Compute the likelihood of the stored data under the hidden Markov model."""
        llh = 0. if log else 1.

        ipsi_dist_evo = self.ext.ipsi.comp_dist_evolution()
        contra_dist_evo = {}
        contra_dist_evo["noext"], contra_dist_evo["ext"] = self.comp_contra_dist_evolution()

        t_stages = self.t_stages if for_t_stage is None else [for_t_stage]
        for stage in t_stages:
            diag_time_matrix = np.diag(self.get_distribution(stage).pmf)
            num_states = ipsi_dist_evo.shape[1]
            marg_joint_state_dist = np.zeros(shape=(num_states, num_states))
            # see the `Bilateral` model for why this is done in this way.
            for case in ["ext", "noext"]:
                joint_state_dist = (
                    ipsi_dist_evo.T
                    @ diag_time_matrix
                    @ contra_dist_evo[case]
                )
                marg_joint_state_dist += joint_state_dist
                _model = getattr(self, case)
                patient_llhs = matrix.fast_trace(
                    _model.ipsi.diagnose_matrices[stage].T,
                    joint_state_dist @ _model.contra.diagnose_matrices[stage]
                )
                llh = add_or_mult(llh, patient_llhs, log=log)

            try:
                marg_patient_llhs = matrix.fast_trace(
                    self.unknown.ipsi.diagnose_matrices[stage].T,
                    marg_joint_state_dist @ self.unknown.contra.diagnose_matrices[stage]
                )
                llh = add_or_mult(llh, marg_patient_llhs, log=log)
            except AttributeError:
                # an AttributeError is raised both when the model has no `unknown`
                # attribute and when no data is loaded in the `unknown` model.
                pass

        if self.use_central:
            if log:
                llh += self.central.likelihood(log=log, for_t_stage=for_t_stage)
            else:
                llh *= self.central.likelihood(log=log, for_t_stage=for_t_stage)

        return llh


    def likelihood(
        self,
        given_params: Iterable[float] | dict[str, float] | None = None,
        log: bool = True,
        mode: Literal["HMM", "BN"] = "HMM",
        for_t_stage: str | None = None,
    ) -> float:
        """Compute the (log-)likelihood of the stored data given the model (and params).

        See the documentation of :py:meth:`.types.Model.likelihood` for more
        information on how to use the ``given_params`` parameter.

        Returns the log-likelihood if ``log`` is set to ``True``. Note that in contrast
        to the :py:class:`.Bilateral` model, the midline model does not support the
        Bayesian network mode.

        Note:
            The computation is faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.

        See Also:
            :py:meth:`.Unilateral.likelihood`
                The corresponding unilateral function.
        """
        try:
            # all functions and methods called here should raise a ValueError if the
            # given parameters are invalid...
            if given_params is None:
                pass
            elif isinstance(given_params, dict):
                self.set_params(**given_params)
            else:
                self.set_params(*given_params)
        except ValueError:
            return -np.inf if log else 0.

        if mode == "HMM":
            return self._hmm_likelihood(log, for_t_stage)

        raise NotImplementedError("Only HMM mode is supported as of now.")


    def risk(
        self,
        involvement: PatternType | None = None,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: dict[str, DiagnoseType] | None = None,
        t_stage: str = "early",
        midline_extension: bool = False,
        central: bool = False,
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> float:
        """Compute the risk of nodal involvement ``given_diagnoses``.

        In addition to the arguments of the :py:meth:`.Bilateral.risk` method, this
        also allows specifying if the patient's tumor extended over the mid-sagittal
        line (``midline_extension=True``) or if it was even located right on that line
        (``central=True``).

        For logical reasons, ``midline_extension=False`` makes no sense if
        ``central=True`` and is thus ignored.
        """
        if given_params is None:
            pass
        elif isinstance(given_params, dict):
            self.set_params(**given_params)
        else:
            self.set_params(*given_params)

        if central:
            return self.central.risk(
                given_diagnoses=given_diagnoses,
                t_stage=t_stage,
                involvement=involvement,
                mode=mode,
            )
        if midline_extension:
            return self.ext.risk(
                given_diagnoses=given_diagnoses,
                t_stage=t_stage,
                involvement=involvement,
                mode=mode,
            )
        return self.noext.risk(
            given_diagnoses=given_diagnoses,
            t_stage=t_stage,
            involvement=involvement,
            mode=mode,
        )


    def draw_patients(
        self,
        num: int,
        stage_dist: Iterable[float],
        rng: np.random.Generator | None = None,
        seed: int = 42,
    ) -> pd.DataFrame:
        """Draw ``num`` patients from the parameterized model."""
        if rng is None:
            rng = np.random.default_rng(seed)

        if sum(stage_dist) != 1.:
            warnings.warn("Sum of stage distribution is not 1. Renormalizing.")
            stage_dist = np.array(stage_dist) / sum(stage_dist)

        if self.use_central:
            raise NotImplementedError(
                "Drawing patients from the central model not yet supported."
            )

        drawn_t_stages = rng.choice(
            a=self.t_stages,
            p=stage_dist,
            size=num,
        )
        distributions = self.get_all_distributions()
        drawn_diag_times = np.array([
            distributions[t_stage].draw_diag_times(rng=rng)
            for t_stage in drawn_t_stages
        ])

        if self.use_midext_evo:
            midext_evo = self.comp_midext_evolution()
            drawn_midexts = np.array([
                rng.choice(a=[False, True], p=midext_evo[t])
                for t in drawn_diag_times
            ])
        else:
            drawn_midexts = rng.choice(
                a=[False, True],
                p=[1. - self.midext_prob, self.midext_prob],
                size=num,
            )

        ipsi_evo = self.ext.ipsi.comp_dist_evolution()
        drawn_diags = np.empty(shape=(num, len(self.ext.ipsi.obs_list)))
        for case in ["ext", "noext"]:
            case_model = getattr(self, case)
            drawn_ipsi_diags = draw_diagnoses(
                diagnose_times=drawn_diag_times[drawn_midexts == (case == "ext")],
                state_evolution=ipsi_evo,
                observation_matrix=case_model.ipsi.observation_matrix(),
                possible_diagnoses=case_model.ipsi.obs_list,
                rng=rng,
                seed=seed,
            )
            drawn_contra_diags = draw_diagnoses(
                diagnose_times=drawn_diag_times[drawn_midexts == (case == "ext")],
                state_evolution=case_model.contra.comp_dist_evolution(),
                observation_matrix=case_model.contra.observation_matrix(),
                possible_diagnoses=case_model.contra.obs_list,
                rng=rng,
                seed=seed,
            )
            drawn_case_diags = np.concatenate([drawn_ipsi_diags, drawn_contra_diags], axis=1)
            drawn_diags[drawn_midexts == (case == "ext")] = drawn_case_diags

        # construct MultiIndex with "ipsi" and "contra" at top level to allow
        # concatenation of the two separate drawn diagnoses
        sides = ["ipsi", "contra"]
        modality_names = list(self.get_all_modalities().keys())
        lnl_names = [lnl for lnl in self.ext.ipsi.graph.lnls.keys()]
        multi_cols = pd.MultiIndex.from_product([sides, modality_names, lnl_names])

        # reorder the column levels and thus also the individual columns to match the
        # LyProX format without mixing up the data
        dataset = pd.DataFrame(drawn_diags, columns=multi_cols)
        dataset = dataset.reorder_levels(order=[1, 0, 2], axis="columns")
        dataset = dataset.sort_index(axis="columns", level=0)
        dataset["tumor", "1", "t_stage"] = drawn_t_stages
        dataset["tumor", "1", "extension"] = drawn_midexts
        dataset["patient", "#", "diagnose_time"] = drawn_diag_times

        return dataset



"""
This package contains code to model the spread of microscopic metastases
through a system of lymph node levels (LNLs), using either a Bayesian network
or a hidden Markov model.
"""
import logging
from logging import NullHandler, StreamHandler

from lymph._version import version

__version__ = version
__description__ = "Package for statistical modelling of lymphatic metastatic spread."
__author__ = "Roman Ludwig"
__email__ = "roman.ludwig@usz.ch"
__uri__ = "https://github.com/rmnldwg/lymph"

# nopycln: file

from lymph import diagnose_times, graph, matrix, models
from lymph.helper import clinical, pathological

__all__ = [
    "diagnose_times", "matrix",
    "graph", "models",
    "clinical", "pathological",
]


# configure library logging akin to how it was done in urllib3 v2.0.4:
# https://github.com/urllib3/urllib3/blob/2.0.4/src/urllib3/__init__.py#L87-L107
logging.getLogger(__name__).addHandler(NullHandler())

def add_stderr_logging(level: int = logging.DEBUG) -> StreamHandler:
    """Add a stderr log handler to the logger."""
    logger = logging.getLogger(__name__)
    handler = StreamHandler()
    handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(level)
    logger.debug("Added stderr logging with level %s.", level)
    return handler



"""Test the binary unilateral system."""
import unittest

import numpy as np

from lymph.graph import LymphNodeLevel, Tumor

from . import fixtures


class InitTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the initialization of a binary model."""

    def test_value_errors(self):
        """Check that the model raises errors when the graph has issues."""
        empty_graph = {}
        only_tumor = {("tumor", "T"): []}
        only_lnl = {("lnl", "II"): []}
        duplicate_lnls = {
            ("tumor", "T"): ["II", "II"],
            ("lnl", "II"): [],
        }
        self.assertRaises(ValueError, lambda: type(self.model)(empty_graph))
        self.assertRaises(ValueError, lambda: type(self.model)(only_tumor))
        self.assertRaises(ValueError, lambda: type(self.model)(only_lnl))
        self.assertRaises(ValueError, lambda: type(self.model)(duplicate_lnls))

    def test_num_nodes(self):
        """Check number of nodes initialized."""
        num_nodes = len(self.graph_dict)
        num_tumor = len({name for kind, name in self.graph_dict if kind == "tumor"})
        num_lnls = len({name for kind, name in self.graph_dict if kind == "lnl"})

        self.assertEqual(len(self.model.graph.nodes), num_nodes)
        self.assertEqual(len(self.model.graph.tumors), num_tumor)
        self.assertEqual(len(self.model.graph.lnls), num_lnls)

    def test_num_edges(self):
        """Check number of edges initialized."""
        num_edges = sum(len(receiving_nodes) for receiving_nodes in self.graph_dict.values())
        num_tumor_edges = sum(
            len(receiving_nodes) for (kind, _), receiving_nodes in self.graph_dict.items()
            if kind == "tumor"
        )
        num_lnl_edges = sum(
            len(receiving_nodes) for (kind, _), receiving_nodes in self.graph_dict.items()
            if kind == "lnl"
        )

        self.assertEqual(len(self.model.graph.edges), num_edges)
        self.assertEqual(len(self.model.graph.tumor_edges), num_tumor_edges)
        self.assertEqual(len(self.model.graph.lnl_edges), num_lnl_edges)
        self.assertEqual(len(self.model.graph.growth_edges), 0)

    def test_tumor(self):
        """Make sure the tumor has been initialized correctly."""
        tumor = self.model.graph.nodes["T"]
        state = tumor.state
        self.assertIsInstance(tumor, Tumor)
        self.assertListEqual(tumor.allowed_states, [state])

    def test_lnls(self):
        """Test they are all binary lymph node levels."""
        model_allowed_states = self.model.graph.allowed_states
        self.assertEqual(len(model_allowed_states), 2)

        for lnl in self.model.graph.lnls.values():
            self.assertIsInstance(lnl, LymphNodeLevel)
            self.assertTrue(lnl.is_binary)
            self.assertEqual(lnl.allowed_states, model_allowed_states)

    def test_tumor_to_lnl_edges(self):
        """Make sure the tumor to LNL edges have been initialized correctly."""
        tumor = self.model.graph.nodes["T"]
        receiving_lnls = self.graph_dict[("tumor", "T")]
        connecting_edge_names = [f"{tumor.name}_to_{lnl}" for lnl in receiving_lnls]

        for edge in self.model.graph.tumor_edges.values():
            self.assertEqual(edge.parent.name, "T")
            self.assertIn(edge.child.name, receiving_lnls)
            self.assertTrue(edge.is_tumor_spread)
            self.assertIn(edge.get_name(middle="_to_"), connecting_edge_names)


class ParameterAssignmentTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the assignment of parameters in a binary model."""

    def test_params_assignment_via_lookup(self):
        """Make sure the spread parameters are assigned correctly."""
        params_to_set = self.create_random_params()
        edges_and_dists = self.model.graph.edges.copy()
        edges_and_dists.update(self.model.get_all_distributions())

        for param_name, value in params_to_set.items():
            name, type_ = param_name.rsplit("_", maxsplit=1)
            edges_and_dists[name].set_params(**{type_: value})
            self.assertEqual(
                edges_and_dists[name].get_params()[type_],
                value,
            )

    def test_params_assignment_via_method(self):
        """Make sure the spread parameters are assigned correctly."""
        params_to_set = self.create_random_params()
        self.model.set_params(**params_to_set)

        edges_and_dists = self.model.graph.edges.copy()
        edges_and_dists.update(self.model.get_all_distributions())

        for param_name, value in params_to_set.items():
            name, type_ = param_name.rsplit("_", maxsplit=1)
            self.assertEqual(
                edges_and_dists[name].get_params()[type_],
                value,
            )

    def test_transition_matrix_deletion(self):
        """Check if the transition matrix gets deleted when a parameter is set."""
        first_lnl_name = list(self.model.graph.lnls.values())[0].name
        trans_mat = self.model.transition_matrix()
        self.model.graph.edges[f"Tto{first_lnl_name}"].set_spread_prob(0.5)
        self.assertFalse(np.all(
            trans_mat == self.model.transition_matrix()
        ))


class TransitionMatrixTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the generation of the transition matrix in a binary model."""

    def setUp(self):
        """Initialize a simple binary model."""
        super().setUp()
        self.model.set_params(**self.create_random_params())

    def test_shape(self):
        """Make sure the transition matrix has the correct shape."""
        num_lnls = len({name for kind, name in self.graph_dict if kind == "lnl"})
        self.assertEqual(self.model.transition_matrix().shape, (2**num_lnls, 2**num_lnls))

    def test_is_probabilistic(self):
        """Make sure the rows of the transition matrix sum to one."""
        row_sums = np.sum(self.model.transition_matrix(), axis=1)
        self.assertTrue(np.allclose(row_sums, 1.))

    @staticmethod
    def is_recusively_upper_triangular(mat: np.ndarray) -> bool:
        """Return `True` is `mat` is recursively upper triangular."""
        if mat.shape == (1, 1):
            return True

        if not np.all(np.equal(np.triu(mat), mat)):
            return False

        half = mat.shape[0] // 2
        for i in [0, 1]:
            for j in [0, 1]:
                return TransitionMatrixTestCase.is_recusively_upper_triangular(
                    mat[i * half:(i + 1) * half, j * half:(j + 1) * half]
                )

    def test_is_recusively_upper_triangular(self) -> None:
        """Make sure the transition matrix is recursively upper triangular."""
        self.assertTrue(self.is_recusively_upper_triangular(self.model.transition_matrix()))


class ObservationMatrixTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the generation of the observation matrix in a binary model."""

    def setUp(self):
        """Initialize a simple binary model."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)

    def test_shape(self):
        """Make sure the observation matrix has the correct shape."""
        num_lnls = len(self.model.graph.lnls)
        num_modalities = len(self.model.get_all_modalities())
        expected_shape = (2**num_lnls, 2**(num_lnls * num_modalities))
        self.assertEqual(self.model.observation_matrix().shape, expected_shape)

    def test_is_probabilistic(self):
        """Make sure the rows of the observation matrix sum to one."""
        row_sums = np.sum(self.model.observation_matrix(), axis=1)
        self.assertTrue(np.allclose(row_sums, 1.))


class PatientDataTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test loading the patient data."""

    def setUp(self):
        """Load patient data."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric", foo="frozen")
        self.model.set_params(**self.create_random_params())
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def test_load_patient_data(self):
        """Make sure the patient data is loaded correctly."""
        self.assertEqual(len(self.model.patient_data), len(self.raw_data))
        self.assertRaises(
            ValueError, self.model.load_patient_data, self.raw_data, side="foo"
        )

    def test_t_stages(self):
        """Make sure all T-stages are present."""
        t_stages_in_data = self.model.get_t_stages("data")
        t_stages_in_diag_time_dists = self.model.get_t_stages("distributions")
        t_stages_in_model = self.model.get_t_stages("valid")
        t_stages_intersection = set(t_stages_in_data).intersection(t_stages_in_diag_time_dists)

        self.assertNotIn("foo", t_stages_in_model)
        self.assertEqual(len(t_stages_in_diag_time_dists), 3)
        self.assertEqual(len(t_stages_intersection), 2)
        self.assertEqual(len(t_stages_intersection), len(t_stages_in_model))

        for t_stage in t_stages_in_model:
            self.assertIn(t_stage, t_stages_in_data)
            self.assertIn(t_stage, t_stages_in_diag_time_dists)

    def test_data_matrices(self):
        """Make sure the data matrices are generated correctly."""
        for t_stage in ["early", "late"]:
            has_t_stage = self.raw_data["tumor", "1", "t_stage"].isin({
                "early": [0,1,2],
                "late": [3,4],
            }[t_stage])
            data_matrix = self.model.data_matrices[t_stage]

            self.assertTrue(t_stage in self.model.data_matrices)
            self.assertEqual(
                data_matrix.shape[0],
                self.model.observation_matrix().shape[1],
            )
            self.assertEqual(
                data_matrix.shape[1],
                has_t_stage.sum(),
            )

    def test_diagnose_matrices(self):
        """Make sure the diagnose matrices are generated correctly."""
        for t_stage in ["early", "late"]:
            has_t_stage = self.raw_data["tumor", "1", "t_stage"].isin({
                "early": [0,1,2],
                "late": [3,4],
            }[t_stage])
            diagnose_matrix = self.model.diagnose_matrices[t_stage]

            self.assertTrue(t_stage in self.model.diagnose_matrices)
            self.assertEqual(
                diagnose_matrix.shape[0],
                self.model.transition_matrix().shape[1],
            )
            self.assertEqual(
                diagnose_matrix.shape[1],
                has_t_stage.sum(),
            )
            # some times, entries in the diagnose matrix are almost one, but just
            # slightly larger. That's why we also have to have the `isclose` here.
            self.assertTrue(np.all(
                np.isclose(diagnose_matrix, 1.)
                | np.less_equal(diagnose_matrix, 1.)
            ))


class LikelihoodTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test the likelihood of a model."""

    def setUp(self):
        """Load patient data."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())
        self.load_patient_data(filename="2021-usz-oropharynx.csv")

    def test_log_likelihood_smaller_zero(self):
        """Make sure the log-likelihood is smaller than zero."""
        likelihood = self.model.likelihood(log=True, mode="HMM")
        self.assertLess(likelihood, 0.)

    def test_likelihood_invalid_params_isinf(self):
        """Make sure the likelihood is `-np.inf` for invalid parameters."""
        random_params = self.create_random_params()
        for name in random_params:
            random_params[name] += 1.
        likelihood = self.model.likelihood(
            given_params=random_params,
            log=True,
            mode="HMM",
        )
        self.assertEqual(likelihood, -np.inf)


class RiskTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Test anything related to the risk computation."""

    def setUp(self):
        """Load params."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())

    def create_random_diagnoses(self):
        """Create a random diagnosis for each modality and LNL."""
        lnl_names = list(self.model.graph.lnls.keys())
        diagnoses = {}

        for modality in self.model.get_all_modalities():
            diagnoses[modality] = fixtures.create_random_pattern(lnl_names)

        return diagnoses

    def test_comp_diagnose_encoding(self):
        """Check computation of one-hot encoding of diagnoses."""
        random_diagnoses = self.create_random_diagnoses()
        num_lnls = len(self.model.graph.lnls)
        num_mods = len(self.model.get_all_modalities())
        num_posible_diagnoses = 2**(num_lnls * num_mods)

        diagnose_encoding = self.model.comp_diagnose_encoding(random_diagnoses)
        self.assertEqual(diagnose_encoding.shape, (num_posible_diagnoses,))
        self.assertEqual(diagnose_encoding.dtype, bool)

    def test_posterior_state_dist(self):
        """Make sure the posterior state dist is correctly computed."""
        posterior_state_dist = self.model.comp_posterior_state_dist(
            given_params=self.create_random_params(),
            given_diagnoses=self.create_random_diagnoses(),
            t_stage=self.rng.choice(["early", "late"]),
        )
        self.assertEqual(posterior_state_dist.shape, (2**len(self.model.graph.lnls),))
        self.assertEqual(posterior_state_dist.dtype, float)
        self.assertTrue(np.isclose(np.sum(posterior_state_dist), 1.))

    def test_risk(self):
        """Make sure the risk is correctly computed."""
        random_pattern = fixtures.create_random_pattern(self.model.graph.lnls.keys())
        random_diagnoses = self.create_random_diagnoses()
        random_t_stage = self.rng.choice(["early", "late"])
        random_params = self.create_random_params()

        risk = self.model.risk(
            involvement=random_pattern,
            given_params=random_params,
            given_diagnoses=random_diagnoses,
            t_stage=random_t_stage,
        )
        self.assertEqual(risk.dtype, float)
        self.assertGreaterEqual(risk, 0.)
        self.assertLessEqual(risk, 1.)


class DataGenerationTestCase(fixtures.BinaryUnilateralModelMixin, unittest.TestCase):
    """Check the data generation utilities."""

    def setUp(self):
        """Load params."""
        super().setUp()
        self.model.replace_all_modalities(fixtures.MODALITIES)
        self.init_diag_time_dists(early="frozen", late="parametric")
        self.model.set_params(**self.create_random_params())

    def test_generate_early_patients(self):
        """Check that generating only early T-stage patients works."""
        early_patients = self.model.draw_patients(
            num=100,
            stage_dist=[1., 0.],
            rng=self.rng,
        )
        self.assertEqual(len(early_patients), 100)
        self.assertEqual(sum(early_patients["tumor", "1", "t_stage"] == "early"), 100)
        self.assertIn(("CT", "ipsi", "II"), early_patients.columns)
        self.assertIn(("FNA", "ipsi", "III"), early_patients.columns)

    def test_generate_late_patients(self):
        """Check that generating only late T-stage patients works."""
        late_patients = self.model.draw_patients(
            num=100,
            stage_dist=[0., 1.],
            rng=self.rng,
        )
        self.assertEqual(len(late_patients), 100)
        self.assertEqual(sum(late_patients["tumor", "1", "t_stage"] == "late"), 100)
        self.assertIn(("CT", "ipsi", "II"), late_patients.columns)
        self.assertIn(("FNA", "ipsi", "III"), late_patients.columns)

    def test_distribution_of_patients(self):
        """Check that the distribution of LNL involvement is correct."""
        # set spread params all to 0
        for lnl_edge in self.model.graph.lnl_edges.values():
            lnl_edge.set_spread_prob(0.)

        # make all patients diagnosed after exactly one time-step
        self.model.set_distribution("early", [0,1,0,0,0,0,0,0,0,0,0])

        # assign only one pathology modality
        self.model.clear_modalities()
        self.model.set_modality("tmp", spec=1., sens=1.)

        # extract the tumor spread parameters
        params = self.model.get_params(as_dict=True)
        params = {
            key.replace("Tto", "").replace("_spread", ""): value
            for key, value in params.items()
            if "Tto" in key
        }

        # draw large enough amount of patients
        patients = self.model.draw_patients(
            num=10000,
            stage_dist=[1., 0.],
            rng=self.rng,
        )

        # check that the distribution of LNL involvement matches tumor spread params
        for lnl, expected_mean in params.items():
            actual_mean = patients[("tmp", "ipsi", lnl)].mean()
            self.assertAlmostEqual(actual_mean, expected_mean, delta=0.02)



"""Unit tests for the Edge class."""
import unittest

import numpy as np

from lymph import graph


class BinaryEdgeTestCase(unittest.TestCase):
    """Tests for the Edge class."""

    def setUp(self) -> None:
        super().setUp()
        parent = graph.LymphNodeLevel("parent")
        child = graph.LymphNodeLevel("child")
        self.was_called = False
        self.edge = graph.Edge(parent, child, callbacks=[self.callback])

    def callback(self) -> None:
        """Callback function for the edge."""
        self.was_called = True

    def test_str(self) -> None:
        """Test the string representation of the edge."""
        self.assertEqual(str(self.edge), "Edge parent to child")

    def test_repr(self) -> None:
        """Test if the edge can be recreated from repr."""
        # pylint: disable=eval-used
        recreated_edge = eval(
            repr(self.edge),
            {
                "Edge": graph.Edge,
                "LymphNodeLevel": graph.LymphNodeLevel,
                "Tumor": graph.Tumor,
            },
        )
        self.assertEqual(self.edge.get_name(), recreated_edge.get_name())
        self.assertEqual(self.edge.parent.name, recreated_edge.parent.name)
        self.assertEqual(self.edge.child.name, recreated_edge.child.name)
        self.assertEqual(self.edge.spread_prob, recreated_edge.spread_prob)
        self.assertEqual(self.edge.micro_mod, recreated_edge.micro_mod)

    def test_callback_on_param_change(self) -> None:
        """Test if the callback function is called."""
        self.edge.spread_prob = 0.5
        self.assertTrue(self.was_called)

    def test_graph_change(self) -> None:
        """Check if the callback also works when parent/child nodes are changed."""
        old_child = self.edge.child
        new_child = graph.LymphNodeLevel("new_child")
        self.edge.child = new_child
        self.assertTrue(self.was_called)
        self.assertNotIn(self.edge, old_child.inc)

    def test_transition_tensor_row_sums(self) -> None:
        """Testing the transition tensor."""
        row_sum = self.edge.transition_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sum, 1.0))


class TrinaryEdgeTestCase(unittest.TestCase):
    """Tests for the Edge class in case the parent and child node are trinary."""

    def setUp(self) -> None:
        super().setUp()
        parent = graph.LymphNodeLevel("parent", allowed_states=[0,1,2])
        child = graph.LymphNodeLevel("child", allowed_states=[0,1,2])
        self.edge = graph.Edge(parent, child)
        self.edge.spread_prob = 0.3
        self.edge.micro_mod = 0.7

        self.growth_edge = graph.Edge(parent, parent)
        self.growth_edge.spread_prob = 0.5

    def test_transition_tensor_row_sums(self) -> None:
        """Testing the transition tensor."""
        row_sum = self.edge.transition_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sum, 1.0))

    def test_growth_transition_tensor_row_sums(self) -> None:
        """Testing the transition tensor."""
        row_sum = self.growth_edge.transition_tensor.sum(axis=2)
        self.assertTrue(np.allclose(row_sum, 1.0))



from __future__ import annotations

import warnings
from functools import cached_property
from itertools import product
from typing import Any, Iterable, Literal

import numpy as np
import pandas as pd

from lymph import diagnose_times, graph, matrix, modalities, types
from lymph.helper import (  # nopycln: import
    add_or_mult,
    dict_to_func,
    draw_diagnoses,
    early_late_mapping,
    flatten,
    get_params_from,
    set_params_for,
    smart_updating_dict_cached_property,
)

warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)


class Unilateral(
    diagnose_times.Composite,
    modalities.Composite,
    types.Model,
):
    """Class that models metastatic progression in a unilateral lymphatic system.

    It does this by representing it as a directed graph (DAG), which is stored in and
    managed by the attribute :py:attr:`~graph`. The progression itself can be modelled
    via hidden Markov models (HMM) or Bayesian networks (BN). In both cases, instances
    of this class allow to calculate the probability of a certain hidden pattern of
    involvement, given an individual diagnosis of a patient.
    """
    def __init__(
        self,
        graph_dict: dict[tuple[str], list[str]],
        tumor_state: int | None = None,
        allowed_states: list[int] | None = None,
        max_time: int = 10,
        **_kwargs,
    ) -> None:
        """Create a new instance of the :py:class:`~Unilateral` class.

        The ``graph_dict`` that represents the lymphatic system should given as a
        dictionary. Its keys are tuples of the form ``("tumor", "<tumor_name>")`` or
        ``("lnl", "<lnl_name>")``. The values are lists of strings that represent the
        names of the nodes that are connected to the node given by the key.

        Note:
            Do make sure the values in the dictionary are of type ``list`` and *not*
            ``set``. Sets do not preserve the order of the elements and thus the order
            of the edges in the graph. This may lead to inconsistencies in the model.

        For example, the following graph represents a lymphatic system with one tumors
        and three lymph node levels:

        .. code-block:: python

            graph = {
                ("tumor", "T"): ["II", "III", "IV"],
                ("lnl", "II"): ["III"],
                ("lnl", "III"): ["IV"],
                ("lnl", "IV"): [],
            }

        The ``tumor_state`` is the initial (and unchangeable) state of the tumor.
        Typically, this can be omitted and is then set to be the maximum of the
        ``allowed_states``, which is the states the LNLs can take on. The default is a
        binary representation with ``allowed_states=[0, 1]``. For this, one can also
        use the classmethod :py:meth:`~Unilateral.binary`. For a trinary representation
        with ``allowed_states=[0, 1, 2]`` use the classmethod
        :py:meth:`~Unilateral.trinary`.

        The ``max_time`` parameter defines the latest possible time step for a
        diagnosis. In the HMM case, the probability disitrubtion over all hidden states
        is evolved from :math:`t=0` to ``max_time``. In the BN case, this parameter has
        no effect.

        The ``is_micro_mod_shared`` and ``is_growth_shared`` parameters determine
        whether the microscopic involvement and growth parameters are shared among all
        LNLs. If they are set to ``True``, the parameters are set globally for all LNLs.
        If they are set to ``False``, the parameters are set individually for each LNL.
        """
        self.graph = graph.Representation(
            graph_dict=graph_dict,
            tumor_state=tumor_state,
            allowed_states=allowed_states,
        )

        diagnose_times.Composite.__init__(self, max_time=max_time, is_distribution_leaf=True)
        modalities.Composite.__init__(self, is_modality_leaf=True)


    @classmethod
    def binary(cls, graph_dict: dict[tuple[str], set[str]], **kwargs) -> Unilateral:
        """Create an instance of the :py:class:`~Unilateral` class with binary LNLs."""
        return cls(graph_dict, allowed_states=[0, 1], **kwargs)


    @classmethod
    def trinary(cls, graph_dict: dict[tuple[str], set[str]], **kwargs) -> Unilateral:
        """Create an instance of the :py:class:`~Unilateral` class with trinary LNLs."""
        return cls(graph_dict, allowed_states=[0, 1, 2], **kwargs)


    def __str__(self) -> str:
        """Print info about the instance."""
        return f"Unilateral with {len(self.graph.tumors)} tumors and {len(self.graph.lnls)} LNLs"


    def print_info(self):
        """Print detailed information about the instance."""
        num_tumors = len(self.graph.tumors)
        num_lnls   = len(self.graph.lnls)
        string = (
            f"Unilateral lymphatic system with {num_tumors} tumor(s) "
            f"and {num_lnls} LNL(s).\n"
            + " ".join([f"{e} {e.spread_prob}%" for e in self.graph.tumor_edges]) + "\n" + " ".join([f"{e} {e.spread_prob}%" for e in self.graph.lnl_edges])
            + f"\n the growth probability is: {self.graph.growth_edges[0].spread_prob}" + f" the micro mod is {self.graph.lnl_edges[0].micro_mod}"
        )
        print(string)


    @property
    def is_trinary(self) -> bool:
        """Return whether the model is trinary."""
        return self.graph.is_trinary

    @property
    def is_binary(self) -> bool:
        """Return whether the model is binary."""
        return self.graph.is_binary


    def get_t_stages(
        self,
        which: Literal["valid", "distributions", "data"] = "valid",
    ) -> list[str]:
        """Return the T-stages of the model."""
        distribution_t_stages = super().t_stages

        try:
            data_t_stages = self.patient_data[("_model", "#", "t_stage")].unique()
        except AttributeError:
            data_t_stages = []

        if which == "valid":
            return sorted(set(distribution_t_stages) & set(data_t_stages))
        if which == "distributions":
            return distribution_t_stages
        if which == "data":
            return data_t_stages

        raise ValueError(
            f"Invalid value for 'which': {which}. Must be either 'valid', "
            "'distributions', or 'data'."
        )


    def get_tumor_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Get the parameters of the tumor spread edges."""
        return get_params_from(self.graph.tumor_edges, as_dict, as_flat)


    def get_lnl_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Get the parameters of the LNL spread edges.

        In the trinary case, this includes the growth parameters as well as the
        microscopic modification parameters.
        """
        return get_params_from(self.graph.lnl_edges, as_dict, as_flat)


    def get_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Get the parameters of the spread edges."""
        return {
            **self.get_tumor_spread_params(as_dict, as_flat),
            **self.get_lnl_spread_params(as_dict, as_flat),
        }


    def get_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Get the parameters of the model.

        If ``as_dict`` is ``True``, the parameters are returned as a dictionary. If
        ``as_flat`` is ``True``, the dictionary is flattened, i.e., all nested
        dictionaries are merged into one, using :py:func:`~lymph.helper.flatten`.
        """
        params = self.get_spread_params(as_flat=as_flat)
        params.update(self.get_distribution_params(as_flat=as_flat))

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def set_tumor_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Assign new parameters to the tumor spread edges."""
        return set_params_for(self.graph.tumor_edges, *args, **kwargs)


    def set_lnl_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Assign new parameters to the LNL spread edges."""
        return set_params_for(self.graph.lnl_edges, *args, **kwargs)


    def set_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Assign new parameters to the spread edges."""
        args = self.set_tumor_spread_params(*args, **kwargs)
        return self.set_lnl_spread_params(*args, **kwargs)


    def set_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Assign new parameters to the model.

        The parameters can be provided either via positional arguments or via keyword
        arguments. The positional arguments are used up one by one first by the
        :py:meth:`lymph.graph.set_params` method and then by the
        :py:meth:`lymph.models.Unilateral.set_distribution_params` method.

        The keyword arguments can be of the format ``"<edge_name>_<param_name>"`` or
        ``"<t_stage>_<param_name>"`` for the distributions over diagnose times. If only
        a ``"<param_name>"`` is provided, it is assumed to be a global parameter and is
        sent to all edges or distributions. But the more specific keyword arguments
        override the global ones, which in turn override the positional arguments.

        >>> graph = {
        ...     ("tumor", "T"): ["II", "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... }
        >>> model = Unilateral.trinary(
        ...     graph_dict=graph,
        ...     is_micro_mod_shared=True,
        ...     is_growth_shared=True,
        ... )
        >>> model.set_params(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.99, AtoB_param="not_used")
        (0.99,)
        >>> model.get_params(as_dict=True)  # doctest: +NORMALIZE_WHITESPACE
        {'TtoII_spread': 0.1,
         'TtoIII_spread': 0.2,
         'II_growth': 0.3,
         'IItoIII_spread': 0.4,
         'IItoIII_micro': 0.5,
         'III_growth': 0.6}
        >>> _ = model.set_params(growth=0.123)
        >>> model.get_params(as_dict=True)  # doctest: +NORMALIZE_WHITESPACE
        {'TtoII_spread': 0.1,
         'TtoIII_spread': 0.2,
         'II_growth': 0.123,
         'IItoIII_spread': 0.4,
         'IItoIII_micro': 0.5,
         'III_growth': 0.123}
        """
        args = self.set_spread_params(*args, **kwargs)
        return self.set_distribution_params(*args, **kwargs)


    def comp_transition_prob(
        self,
        newstate: list[int],
        assign: bool = False
    ) -> float:
        """Computes probability to transition to ``newstate``, given its current state.

        The probability is computed as the product of the transition probabilities of
        the individual LNLs. If ``assign`` is ``True``, the new state is assigned to
        the model using the method :py:meth:`lymph.graph.Representation.set_state`.
        """
        trans_prob = 1
        for i, lnl in enumerate(self.graph.lnls):
            trans_prob *= lnl.comp_trans_prob(new_state = newstate[i])
            if trans_prob == 0:
                break

        if assign:
            self.graph.set_state(newstate)

        return trans_prob


    def comp_diagnose_prob(
        self,
        diagnoses: pd.Series | dict[str, dict[str, bool]]
    ) -> float:
        """Compute the probability to observe a diagnose given the current state.

        The ``diagnoses`` is either a pandas ``Series`` object corresponding to one row
        of a patient data table, or a dictionary with keys of diagnostic modalities and
        values of dictionaries holding the observation for each LNL under the
        respective key.

        It returns the probability of observing this particular combination of
        diagnoses, given the current state of the system.
        """
        prob = 1.
        for name, modality in self.get_all_modalities().items():
            if name in diagnoses:
                mod_diagnose = diagnoses[name]
                for lnl in self.graph.lnls:
                    try:
                        lnl_diagnose = mod_diagnose[lnl.name]
                    except KeyError:
                        continue
                    except IndexError as idx_err:
                        raise ValueError(
                            "diagnoses were not provided in the correct format"
                        ) from idx_err
                    prob *= lnl.comp_obs_prob(lnl_diagnose, modality.confusion_matrix)
        return prob


    @property
    def obs_list(self):
        """Return the list of all possible observations.

        They are ordered the same way as the :py:attr:`.graph.Representation.state_list`,
        but additionally by modality. E.g., for two LNLs II, III and two modalities CT,
        pathology, the list would look like this:

        >>> model = Unilateral(graph_dict={
        ...     ("tumor", "T"): ["II" , "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> model.set_modality("CT", spec=0.8, sens=0.8)
        >>> model.set_modality("pathology", spec=1.0, sens=1.0)
        >>> model.obs_list  # doctest: +ELLIPSIS
        array([[0, 0, 0, 0],
               [0, 0, 0, 1],
               [0, 0, 1, 0],
               [0, 0, 1, 1],
               ...
               [1, 1, 0, 1],
               [1, 1, 1, 0],
               [1, 1, 1, 1]])

        The first two columns correspond to the observation of LNLs II and III under
        modality CT, the second two columns correspond to the same LNLs under the
        pathology modality.
        """
        possible_obs_list = []
        for modality in self.get_all_modalities().values():
            possible_obs = np.arange(modality.confusion_matrix.shape[1])
            for _ in self.graph.lnls:
                possible_obs_list.append(possible_obs.copy())

        return np.array(list(product(*possible_obs_list)))


    def transition_matrix(self) -> np.ndarray:
        """Matrix encoding the probabilities to transition from one state to another.

        This is the crucial object for modelling the evolution of the probabilistic
        system in the context of the hidden Markov model. It has the shape
        :math:`2^N \\times 2^N` where :math:`N` is the number of nodes in the graph.
        The :math:`i`-th row and :math:`j`-th column encodes the probability to
        transition from the :math:`i`-th state to the :math:`j`-th state. The states
        are ordered as in the :py:attr:`.graph.Representation.state_list`.

        See Also:
            :py:func:`~lymph.descriptors.matrix.generate_transition`
                The function actually computing the transition matrix.

        >>> model = Unilateral(graph_dict={
        ...     ("tumor", "T"): ["II", "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> model.set_params(0.7, 0.3, 0.2)  # doctest: +ELLIPSIS
        ()
        >>> model.transition_matrix()
        array([[0.21, 0.09, 0.49, 0.21],
               [0.  , 0.3 , 0.  , 0.7 ],
               [0.  , 0.  , 0.56, 0.44],
               [0.  , 0.  , 0.  , 1.  ]])
        """
        return matrix.generate_transition(
            lnls=self.graph.lnls.values(),
            num_states=3 if self.is_trinary else 2,
        )


    def observation_matrix(self) -> np.ndarray:
        """The matrix encoding the probabilities to observe a certain diagnosis.

        Every element in this matrix holds a probability to observe a certain diagnosis
        (or combination of diagnoses, when using multiple diagnostic modalities) given
        the current state of the system. It has the shape
        :math:`2^N \\times 2^\\{N \\times M\\}` where :math:`N` is the number of nodes in
        the graph and :math:`M` is the number of diagnostic modalities.

        See Also:
            :py:func:`~lymph.descriptors.matrix.generate_observation`
                The function actually computing the observation matrix.
        """
        return matrix.generate_observation(
            modalities=self.get_all_modalities().values(),
            num_lnls=len(self.graph.lnls),
            base=3 if self.is_trinary else 2,
        )


    @smart_updating_dict_cached_property
    def data_matrices(self) -> matrix.DataEncodingUserDict:
        """Holds the data encoding in matrix form for every T-stage.

        See Also:
            :py:class:`~lymph.descriptors.matrix.DataEncodingUserDict`
        """
        return matrix.DataEncodingUserDict(model=self)


    @smart_updating_dict_cached_property
    def diagnose_matrices(self) -> matrix.DiagnoseUserDict:
        """Holds the probability of a patient's diagnosis, given any hidden state.

        Essentially, this is just the data encoding matrix of a certain T-stage
        multiplied with the observation matrix. It is thus also a dictionary with
        keys of T-stages and values of matrices.

        See Also:
            :py:class:`~lymph.descriptors.matrix.DiagnoseUserDict`
        """
        return matrix.DiagnoseUserDict(model=self)


    def load_patient_data(
        self,
        patient_data: pd.DataFrame,
        side: str = "ipsi",
        mapping: callable | dict[int, Any] = early_late_mapping,
    ) -> None:
        """Load patient data in `LyProX`_ format into the model.

        Since the `LyProX`_ data format contains information on both sides (i.e.,
        ipsi- and contralateral) of the neck, the ``side`` parameter is used to select
        the for which of the two to store the involvement data.

        With the ``mapping`` function or dictionary, the reported T-stages (usually 0,
        1, 2, 3, and 4) can be mapped to any keys also used to access the corresponding
        distribution over diagnose times. The default mapping is to map 0, 1, and 2 to
        "early" and 3 and 4 to "late".

        What this method essentially does is to copy the entire data frame, check all
        necessary information is present, and add a new top-level header ``"_model"`` to
        the data frame. Under this header, columns are assembled that contain all the
        information necessary to compute the observation and diagnose matrices.

        .. _LyProX: https://lyprox.org/
        """
        patient_data = patient_data.copy()

        if isinstance(mapping, dict):
            mapping = dict_to_func(mapping)

        for modality_name in self.get_all_modalities().keys():
            if modality_name not in patient_data:
                raise ValueError(f"Modality '{modality_name}' not found in data.")

            if side not in patient_data[modality_name]:
                raise ValueError(f"{side}lateral involvement data not found.")

            for name in self.graph.lnls.keys():
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=pd.errors.PerformanceWarning)
                    modality_side_data = patient_data[modality_name, side]

                if name not in modality_side_data:
                    raise ValueError(f"Involvement data for LNL {name} not found.")
                column = patient_data[modality_name, side, name]
                patient_data["_model", modality_name, name] = column

        patient_data["_model", "#", "t_stage"] = patient_data.apply(
            lambda row: mapping(row["tumor", "1", "t_stage"]), axis=1
        )

        for t_stage in self.get_t_stages("distributions"):
            if t_stage not in patient_data["_model", "#", "t_stage"].values:
                warnings.warn(f"No data for T-stage {t_stage} found.")

        self._patient_data = patient_data
        # Changes to the patient data require a recomputation of the data and
        # diagnose matrices. For the data matrix, it is enough to clear the respective
        # ``UserDict``. For the diagnose matrices, we need to delete the hash value of
        # the patient data, so that the next time it is requested, a cache miss occurs
        # and they are recomputed.
        self.data_matrices.clear()
        try:
            del self.patient_data_hash
        except AttributeError:
            pass


    @cached_property
    def patient_data_hash(self) -> int:
        """Hash of the patient data.

        This is used to check if the patient data has changed since the last time
        the data and diagnose matrices were computed. If so, they are recomputed.
        """
        return hash(self.patient_data.to_numpy().tobytes())


    @property
    def patient_data(self) -> pd.DataFrame:
        """Return the patient data loaded into the model.

        After succesfully loading the data with the method :py:meth:`.load_patient_data`,
        the copied patient data now contains the additional top-level header
        ``"_model"``. Under it, the observed per LNL involvement is listed for every
        diagnostic modality in the dictionary returned by :py:meth:`.get_all_modalities`
        and for each of the LNLs in the list :py:attr:`.graph.Representation.lnls`.

        It also contains information on the patient's T-stage under the header
        ``("_model", "#", "t_stage")``.
        """
        if not hasattr(self, "_patient_data"):
            raise AttributeError("No patient data loaded yet.")

        return self._patient_data


    def evolve_dist(self, state_dist: np.ndarray, num_steps: int) -> np.ndarray:
        """Evolve the ``state_dist`` of possible states over ``num_steps``.

        This is done by multiplying the ``state_dist`` with the transition matrix
        from the left ``num_steps`` times. The result is a new distribution over
        possible states at a new time-step :math:`t' = t + n`, where :math:`n`
        is the number of steps ``num_steps``.
        """
        for _ in range(num_steps):
            state_dist = state_dist @ self.transition_matrix()

        return state_dist


    def comp_dist_evolution(self) -> np.ndarray:
        """Compute a complete evolution of the model.

        This returns a matrix with the distribution over the possible states for
        each time step from :math:`t = 0` to :math:`t = T`, where :math:`T` is the
        maximum diagnose time stored in the model's attribute ``max_time``.

        Note that at this point, the distributions are not weighted with the
        distribution over diagnose times that are stored and managed for each T-stage
        in the dictionary returned by :py:meth:`.get_all_distributions`.
        """
        state_dists = np.zeros(shape=(self.max_time + 1, len(self.graph.state_list)))
        state_dists[0, 0] = 1.

        for t in range(1, self.max_time + 1):
            state_dists[t] = self.evolve_dist(state_dists[t-1], num_steps=1)

        return state_dists


    def comp_state_dist(self, t_stage: str = "early", mode: Literal["HMM", "BN"] = "HMM") -> np.ndarray:
        """Compute the distribution over possible states.

        Do this either for a given ``t_stage``, when ``mode`` is set to ``"HMM"``,
        which is essentially a marginalization of the evolution over the possible
        states as computed by :py:meth:`.comp_dist_evolution` with the distribution
        over diagnose times for the given T-stage from the dictionary returned by
        :py:meth:`.get_all_dsitributions`.

        Or, when ``mode`` is set to ``"BN"``, compute the distribution over states for
        the Bayesian network. In that case, the ``t_stage`` parameter is ignored.
        """
        if mode == "HMM":
            state_dists = self.comp_dist_evolution()
            diag_time_dist = self.get_distribution(t_stage).pmf

            return diag_time_dist @ state_dists

        if mode == "BN":
            state_dist = np.ones(shape=(len(self.graph.state_list),), dtype=float)

            for i, state in enumerate(self.graph.state_list):
                self.graph.set_state(*state)
                for node in self.graph.lnls.values():
                    state_dist[i] *= node.comp_bayes_net_prob()

            return state_dist


    def comp_obs_dist(self, t_stage: str = "early", mode: Literal["HMM", "BN"] = "HMM") -> np.ndarray:
        """Compute the distribution over all possible observations for a given T-stage.

        Returns an array of probabilities for each possible complete observation. This
        entails multiplying the distribution over states as returned by the
        :py:meth:`~comp_state_dist` method with the :py:attr:`~observation_matrix`.

        Note that since the :py:attr:`~observation_matrix` can become very large, this
        method is not very efficient for inference. Instead, we compute the
        :py:attr:`~diagnose_matrices` from the :py:attr:`~observation_matrix` and
        the :py:attr:`~data_matrices` and use these to compute the likelihood.
        """
        state_dist = self.comp_state_dist(t_stage=t_stage, mode=mode)
        return state_dist @ self.observation_matrix()


    def _bn_likelihood(self, log: bool = True, t_stage: str | None = None) -> float:
        """Compute the BN likelihood, using the stored params."""
        if t_stage is None:
            t_stage = "_BN"

        state_dist = self.comp_state_dist(mode="BN")
        patient_llhs = state_dist @ self.diagnose_matrices[t_stage]

        return np.sum(np.log(patient_llhs)) if log else np.prod(patient_llhs)


    def _hmm_likelihood(self, log: bool = True, t_stage: str | None = None) -> float:
        """Compute the HMM likelihood, using the stored params."""
        evolved_model = self.comp_dist_evolution()
        llh = 0. if log else 1.

        if t_stage is None:
            t_stages = self.get_t_stages("valid")
        else:
            t_stages = [t_stage]

        for t_stage in t_stages:
            patient_llhs = (
                self.get_distribution(t_stage).pmf
                @ evolved_model
                @ self.diagnose_matrices[t_stage]
            )
            llh = add_or_mult(llh, patient_llhs, log)

        return llh


    def likelihood(
        self,
        given_params: Iterable[float] | dict[str, float] | None = None,
        log: bool = True,
        mode: Literal["HMM", "BN"] = "HMM",
        for_t_stage: str | None = None,
    ) -> float:
        """Compute the (log-)likelihood of the stored data given the model (and params).

        See the documentation of :py:meth:`lymph.types.Model.likelihood` for more
        information on how to use the ``given_params`` parameter.

        Returns the log-likelihood if ``log`` is set to ``True``. The ``mode`` parameter
        determines whether the likelihood is computed for the hidden Markov model
        (``"HMM"``) or the Bayesian network (``"BN"``).
        """
        try:
            # all functions and methods called here should raise a ValueError if the
            # given parameters are invalid...
            if given_params is None:
                pass
            elif isinstance(given_params, dict):
                self.set_params(**given_params)
            else:
                self.set_params(*given_params)
        except ValueError:
            return -np.inf if log else 0.

        if mode == "HMM":
            return self._hmm_likelihood(log, for_t_stage)
        if mode == "BN":
            return self._bn_likelihood(log, for_t_stage)

        raise ValueError("Invalid mode. Must be either 'HMM' or 'BN'.")


    def comp_diagnose_encoding(
        self,
        given_diagnoses: types.DiagnoseType | None = None,
    ) -> np.ndarray:
        """Compute one-hot vector encoding of a given diagnosis."""
        diagnose_encoding = np.array([True], dtype=bool)

        for modality in self.get_all_modalities().keys():
            diagnose_encoding = np.kron(
                diagnose_encoding,
                matrix.compute_encoding(
                    lnls=self.graph.lnls.keys(),
                    pattern=given_diagnoses.get(modality, {}),
                    base=2,   # diagnoses are always binary!
                ),
            )

        return diagnose_encoding


    def comp_posterior_state_dist(
        self,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: types.DiagnoseType | None = None,
        t_stage: str | int = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> np.ndarray:
        """Compute the posterior distribution over hidden states given a diagnosis.

        The ``given_diagnoses`` is a dictionary of diagnoses for each modality. E.g.,
        this could look like this:

        .. code-block:: python

            given_diagnoses = {
                "MRI": {"II": True, "III": False, "IV": False},
                "PET": {"II": True, "III": True, "IV": None},
            }

        The ``t_stage`` parameter determines the T-stage for which the posterior is
        computed. The ``mode`` parameter determines whether the posterior is computed
        for the hidden Markov model (``"HMM"``) or the Bayesian network (``"BN"``).
        In case of the Bayesian network mode, the ``t_stage`` parameter is ignored.

        Note:
            The computation is much faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.
        """
        # in contrast to when computing the likelihood, we do want to raise an error
        # here if the parameters are invalid, since we want to know if the user
        # provided invalid parameters. In the likelihood, we rather return a zero
        # likelihood to tell the inference algorithm that the parameters are invalid.
        if given_params is None:
            pass
        elif isinstance(given_params, dict):
            self.set_params(**given_params)
        else:
            self.set_params(*given_params)

        if given_diagnoses is None:
            given_diagnoses = {}

        diagnose_encoding = self.comp_diagnose_encoding(given_diagnoses)
        # vector containing P(Z=z|X). Essentially a data matrix for one patient
        diagnose_given_state = diagnose_encoding @ self.observation_matrix().T

        # vector P(X=x) of probabilities of arriving in state x (marginalized over time)
        state_dist = self.comp_state_dist(t_stage, mode=mode)

        # multiply P(Z=z|X) * P(X) elementwise to get vector of joint probs P(Z=z,X)
        joint_diagnose_and_state = state_dist * diagnose_given_state

        # compute vector of probabilities for all possible involvements given the
        # specified diagnosis P(X|Z=z) = P(Z=z,X) / P(X), where P(X) = sum_z P(Z=z,X)
        return joint_diagnose_and_state / np.sum(joint_diagnose_and_state)


    def risk(
        self,
        involvement: types.PatternType | None = None,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: dict[str, types.PatternType] | None = None,
        t_stage: str = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> float | np.ndarray:
        """Compute risk of a certain involvement, given a patient's diagnosis.

        If an ``involvement`` pattern of interest is provided, this method computes
        the risk of seeing just that pattern for the set of given parameters and a
        dictionary of diagnoses for each modality.

        Using the ``mode`` parameter, the risk can be computed either for the hidden
        Markov model (``"HMM"``) or the Bayesian network (``"BN"``). In case of the
        Bayesian network mode, the ``t_stage`` parameter is ignored.

        Note:
            The computation is much faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.

        See Also:
            :py:meth:`comp_posterior_state_dist`
        """
        posterior_state_dist = self.comp_posterior_state_dist(
            given_params, given_diagnoses, t_stage, mode,
        )

        if involvement is None:
            return posterior_state_dist

        # if a specific involvement of interest is provided, marginalize the
        # resulting vector of hidden states to match that involvement of
        # interest
        marginalize_over_states = matrix.compute_encoding(
            lnls=self.graph.lnls.keys(),
            pattern=involvement,
            base=3 if self.is_trinary else 2,
        )
        return marginalize_over_states @ posterior_state_dist


    def draw_diagnoses(
        self,
        diag_times: list[int],
        rng: np.random.Generator | None = None,
        seed: int = 42,
    ) -> np.ndarray:
        """Given some ``diag_times``, draw diagnoses for each LNL.

        >>> model = Unilateral(graph_dict={
        ...     ("tumor", "T"): ["II" , "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> model.set_modality("CT", spec=0.8, sens=0.8)
        >>> model.draw_diagnoses([0, 1, 2, 3, 4])       # doctest: +NORMALIZE_WHITESPACE
        array([[False,  True],
               [False, False],
               [ True, False],
               [False,  True],
               [False, False]])
        >>> draw_diagnoses(                   # this is the same as the previous example
        ...     diagnose_times=[0, 1, 2, 3, 4],
        ...     state_evolution=model.comp_dist_evolution(),
        ...     observation_matrix=model.observation_matrix(),
        ...     possible_diagnoses=model.obs_list,
        ... )
        array([[False,  True],
               [False, False],
               [ True, False],
               [False,  True],
               [False, False]])
        """
        if rng is None:
            rng = np.random.default_rng(seed)

        state_probs_given_time = self.comp_dist_evolution()[diag_times]
        obs_probs_given_time = state_probs_given_time @ self.observation_matrix()

        obs_indices = np.arange(len(self.obs_list))
        drawn_obs_idx = [
            rng.choice(obs_indices, p=obs_prob)
            for obs_prob in obs_probs_given_time
        ]

        return self.obs_list[drawn_obs_idx].astype(bool)


    def draw_patients(
        self,
        num: int,
        stage_dist: Iterable[float],
        rng: np.random.Generator | None = None,
        seed: int = 42,
        **_kwargs,
    ) -> pd.DataFrame:
        """Draw ``num`` random patients from the model.

        For this, a ``stage_dist``, i.e., a distribution over the T-stages, needs to
        be defined. This must be an iterable of probabilities with as many elements as
        there are defined T-stages in the model (accessible via
        :py:meth:`.get_all_distributions`).

        A random number generator can be provided as ``rng``. If ``None``, a new one
        is initialized with the given ``seed`` (or ``42``, by default).

        See Also:
            :py:meth:`lymph.diagnose_times.Distribution.draw_diag_times`
                Method to draw diagnose times from a distribution.
            :py:meth:`lymph.models.Unilateral.draw_diagnoses`
                Method to draw individual diagnoses.
            :py:meth:`lymph.models.Bilateral.draw_patients`
                The corresponding bilateral method.
        """
        if rng is None:
            rng = np.random.default_rng(seed)

        if sum(stage_dist) != 1.:
            warnings.warn("Sum of stage distribution is not 1. Renormalizing.")
            stage_dist = np.array(stage_dist) / sum(stage_dist)

        drawn_t_stages = rng.choice(
            a=self.get_t_stages("distributions"),
            p=stage_dist,
            size=num,
        )
        distributions = self.get_all_distributions()
        drawn_diag_times = [
            distributions[t_stage].draw_diag_times(rng=rng)
            for t_stage in drawn_t_stages
        ]

        drawn_obs = self.draw_diagnoses(drawn_diag_times, rng=rng)

        modality_names = list(self.get_all_modalities().keys())
        lnl_names = list(self.graph.lnls.keys())
        multi_cols = pd.MultiIndex.from_product([modality_names, ["ipsi"], lnl_names])

        dataset = pd.DataFrame(drawn_obs, columns=multi_cols)
        dataset[("tumor", "1", "t_stage")] = drawn_t_stages

        return dataset



from __future__ import annotations

import logging
import warnings
from typing import Any, Iterable, Literal

import numpy as np
import pandas as pd

from lymph import diagnose_times, matrix, modalities, models, types
from lymph.helper import (
    add_or_mult,
    early_late_mapping,
    flatten,
    synchronize_params,
    unflatten_and_split,
)

warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)
logger = logging.getLogger(__name__)


class Bilateral(
    diagnose_times.Composite,
    modalities.Composite,
    types.Model,
):
    """Class that models metastatic progression in a bilateral lymphatic system.

    This is achieved by creating two instances of the
    :py:class:`~lymph.models.Unilateral` model, one for the ipsi- and one for the
    contralateral side of the neck. The two sides are assumed to be independent of each
    other, given the diagnose time over which we marginalize.

    See Also:
        :py:class:`~lymph.models.Unilateral`
            Two instances of this class are created as attributes. One for the ipsi- and
            one for the contralateral side of the neck.
    """
    def __init__(
        self,
        graph_dict: dict[tuple[str], list[str]],
        is_symmetric: dict[str, bool] | None = None,
        unilateral_kwargs: dict[str, Any] | None = None,
        ipsilateral_kwargs: dict[str, Any] | None = None,
        contralateral_kwargs: dict[str, Any] | None = None,
        **_kwargs,
    ) -> None:
        """Initialize both sides of the neck as :py:class:`.models.Unilateral`.

        The ``graph_dict`` is a dictionary of tuples as keys and lists of strings as
        values. It is passed to both :py:class:`.models.Unilateral` instances,
        which in turn pass it to the :py:class:`.graph.Representation` class that
        stores the graph.

        With the dictionary ``is_symmetric`` the user can specify which aspects of the
        model are symmetric. Valid keys are ``"tumor_spread"`` and ``"lnl_spread"``.
        The values are booleans, with ``True`` meaning that the aspect is symmetric.

        Note:
            The symmetries of tumor and LNL spread are only guaranteed if the
            respective parameters are set via the :py:meth:`.set_params()` method of
            this bilateral model. It is still possible to set different parameters for
            the ipsi- and contralateral side by using their respective
            :py:meth:`.Unilateral.set_params()` method.

        The ``unilateral_kwargs`` are passed to both instances of the unilateral model,
        while the ``ipsilateral_kwargs`` and ``contralateral_kwargs`` are passed to the
        ipsi- and contralateral side, respectively. The ipsi- and contralateral kwargs
        override the unilateral kwargs and may also override the ``graph_dict``. This
        allows the user to specify different graphs for the two sides of the neck.
        """
        self._init_models(
            graph_dict=graph_dict,
            unilateral_kwargs=unilateral_kwargs,
            ipsilateral_kwargs=ipsilateral_kwargs,
            contralateral_kwargs=contralateral_kwargs,
        )

        if is_symmetric is None:
            is_symmetric = {}

        is_symmetric["tumor_spread"] = is_symmetric.get("tumor_spread", False)
        is_symmetric["lnl_spread"] = is_symmetric.get("lnl_spread", True)

        self.is_symmetric = is_symmetric

        diagnose_times.Composite.__init__(
            self,
            distribution_children={"ipsi": self.ipsi, "contra": self.contra},
            is_distribution_leaf=False,
        )
        modalities.Composite.__init__(
            self,
            modality_children={"ipsi": self.ipsi, "contra": self.contra},
            is_modality_leaf=False,
        )


    def _init_models(
        self,
        graph_dict: dict[tuple[str], list[str]],
        unilateral_kwargs: dict[str, Any] | None = None,
        ipsilateral_kwargs: dict[str, Any] | None = None,
        contralateral_kwargs: dict[str, Any] | None = None,
    ):
        """Initialize the two unilateral models."""
        if unilateral_kwargs is None:
            unilateral_kwargs = {}

        ipsi_kwargs = unilateral_kwargs.copy()
        ipsi_kwargs["graph_dict"] = graph_dict
        ipsi_kwargs.update(ipsilateral_kwargs or {})

        contra_kwargs = unilateral_kwargs.copy()
        contra_kwargs["graph_dict"] = graph_dict
        contra_kwargs.update(contralateral_kwargs or {})

        self.ipsi   = models.Unilateral(**ipsi_kwargs)
        self.contra = models.Unilateral(**contra_kwargs)


    @classmethod
    def binary(cls, *args, **kwargs) -> Bilateral:
        """Initialize a binary bilateral model.

        This is a convenience method that sets the ``allowed_states`` of the
        ``unilateral_kwargs`` to ``[0, 1]``. All other ``args`` and ``kwargs`` are
        passed to the :py:meth:`.__init__` method.
        """
        unilateral_kwargs = kwargs.pop("unilateral_kwargs", {})
        unilateral_kwargs["allowed_states"] = [0, 1]
        return cls(*args, unilateral_kwargs=unilateral_kwargs, **kwargs)

    @classmethod
    def trinary(cls, *args, **kwargs) -> Bilateral:
        """Initialize a trinary bilateral model.

        This is a convenience method that sets the ``allowed_states`` of the
        ``unilateral_kwargs`` to ``[0, 1, 2]``. All other ``args`` and ``kwargs`` are
        passed to the :py:meth:`.__init__` method.
        """
        unilateral_kwargs = kwargs.pop("unilateral_kwargs", {})
        unilateral_kwargs["allowed_states"] = [0, 1, 2]
        return cls(*args, unilateral_kwargs=unilateral_kwargs, **kwargs)


    @property
    def is_trinary(self) -> bool:
        """Return whether the model is trinary."""
        if self.ipsi.is_trinary != self.contra.is_trinary:
            raise ValueError("Both sides must be of the same 'naryity'.")

        return self.ipsi.is_trinary

    @property
    def is_binary(self) -> bool:
        """Return whether the model is binary."""
        if self.ipsi.is_binary != self.contra.is_binary:
            raise ValueError("Both sides must be of the same 'naryity'.")

        return self.ipsi.is_binary


    def get_tumor_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the model's spread from tumor to LNLs.

        If the attribute dictionary :py:attr:`.is_symmetric` stores the key-value pair
        ``"tumor_spread": True``, the parameters are returned as a single dictionary,
        since they are the same ipsi- and contralaterally. Otherwise, the parameters
        are returned as a dictionary with two keys, ``"ipsi"`` and ``"contra"``.
        """
        params = {
            "ipsi": self.ipsi.get_tumor_spread_params(as_flat=as_flat),
            "contra": self.contra.get_tumor_spread_params(as_flat=as_flat),
        }

        if self.is_symmetric["tumor_spread"]:
            if params["ipsi"] != params["contra"]:
                warnings.warn(
                    "The tumor spread parameters are not symmetric. "
                    "Returning the ipsilateral parameters."
                )

            params = params["ipsi"]

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def get_lnl_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the model's spread from LNLs to tumor.

        Similarily to the :py:meth:`.get_tumor_spread_params` method, this returns only
        one dictionary if the attribute dictionary :py:attr:`.is_symmetric` stores the
        key-value pair ``"lnl_spread": True``. Otherwise, the parameters are returned
        as a dictionary with two keys, ``"ipsi"`` and ``"contra"``.
        """
        params = {
            "ipsi": self.ipsi.get_lnl_spread_params(as_flat=as_flat),
            "contra": self.contra.get_lnl_spread_params(as_flat=as_flat),
        }

        if self.is_symmetric["lnl_spread"]:
            if params["ipsi"] != params["contra"]:
                warnings.warn(
                    "The LNL spread parameters are not symmetric. "
                    "Returning the ipsilateral parameters."
                )

            params = params["ipsi"]

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def get_spread_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the model's spread edges.

        Depending on the symmetries (i.e. the ``is_symmetric`` attribute), this returns
        different results:

        If ``is_symmetric["tumor_spread"] = False``, the flattened (``as_flat=True``)
        dictionary (``as_dict=True``) will contain keys of the form
        ``ipsi_Tto<lnl>_spread`` and ``contra_Tto<lnl>_spread``, where ``<lnl>`` is the
        name of the lymph node level. However, if the tumor spread is set to be
        symmetric, the leading ``ipsi_`` or ``contra_`` is omitted, since it's valid
        for both sides.

        This is consistent with how the :py:meth:`~lymph.models.Bilteral.set_params`
        method expects the keyword arguments in case of the symmetry configurations.

        >>> model = Bilateral(graph_dict={
        ...     ("tumor", "T"): ["II", "III"],
        ...     ("lnl", "II"): ["III"],
        ...     ("lnl", "III"): [],
        ... })
        >>> num_dims = model.get_num_dims()
        >>> model.set_spread_params(*np.round(np.linspace(0., 1., num_dims+1), 2))
        (1.0,)
        >>> model.get_spread_params(as_flat=False)   # doctest: +NORMALIZE_WHITESPACE
        {'ipsi':    {'TtoII': {'spread': 0.0},
                     'TtoIII': {'spread': 0.2}},
         'contra':  {'TtoII': {'spread': 0.4},
                     'TtoIII': {'spread': 0.6}},
         'IItoIII': {'spread': 0.8}}
        >>> model.get_spread_params(as_flat=True)    # doctest: +NORMALIZE_WHITESPACE
        {'ipsi_TtoII_spread': 0.0,
         'ipsi_TtoIII_spread': 0.2,
         'contra_TtoII_spread': 0.4,
         'contra_TtoIII_spread': 0.6,
         'IItoIII_spread': 0.8}
        """
        params = self.get_tumor_spread_params(as_flat=False)

        if not self.is_symmetric["tumor_spread"] and not self.is_symmetric["lnl_spread"]:
            params["ipsi"].update(self.get_lnl_spread_params(as_flat=False)["ipsi"])
            params["contra"].update(self.get_lnl_spread_params(as_flat=False)["contra"])
        else:
            params.update(self.get_lnl_spread_params(as_flat=as_flat))

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def get_params(
        self,
        as_dict: bool = True,
        as_flat: bool = True,
    ) -> Iterable[float] | dict[str, float]:
        """Return the parameters of the model.

        It returns the combination of the call to the :py:meth:`.Unilateral.get_params`
        of the ipsi- and contralateral side. For the use of the ``as_dict`` and
        ``as_flat`` arguments, see the documentation of the
        :py:meth:`.types.Model.get_params` method.

        Also see the :py:meth:`.get_spread_params` method to understand how the
        symmetry settings affect the return value.
        """
        params = self.get_spread_params(as_flat=as_flat)
        params.update(self.get_distribution_params(as_flat=as_flat))

        if as_flat or not as_dict:
            params = flatten(params)

        return params if as_dict else params.values()


    def set_tumor_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Set the parameters of the model's spread from tumor to LNLs."""
        kwargs, global_kwargs = unflatten_and_split(kwargs, expected_keys=["ipsi", "contra"])

        ipsi_kwargs = global_kwargs.copy()
        ipsi_kwargs.update(kwargs.get("ipsi", {}))
        contra_kwargs = global_kwargs.copy()
        contra_kwargs.update(kwargs.get("contra", {}))

        args = self.ipsi.set_tumor_spread_params(*args, **ipsi_kwargs)
        if self.is_symmetric["tumor_spread"]:
            synchronize_params(
                get_from=self.ipsi.graph.tumor_edges,
                set_to=self.contra.graph.tumor_edges,
            )
        else:
            args = self.contra.set_tumor_spread_params(*args, **contra_kwargs)

        return args


    def set_lnl_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Set the parameters of the model's spread from LNLs to tumor."""
        kwargs, global_kwargs = unflatten_and_split(kwargs, expected_keys=["ipsi", "contra"])

        ipsi_kwargs = global_kwargs.copy()
        ipsi_kwargs.update(kwargs.get("ipsi", {}))
        contra_kwargs = global_kwargs.copy()
        contra_kwargs.update(kwargs.get("contra", {}))

        args = self.ipsi.set_lnl_spread_params(*args, **ipsi_kwargs)
        if self.is_symmetric["lnl_spread"]:
            synchronize_params(
                get_from=self.ipsi.graph.lnl_edges,
                set_to=self.contra.graph.lnl_edges,
            )
        else:
            args = self.contra.set_lnl_spread_params(*args, **contra_kwargs)

        return args


    def set_spread_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Set the parameters of the model's spread edges."""
        args = self.set_tumor_spread_params(*args, **kwargs)
        return self.set_lnl_spread_params(*args, **kwargs)


    def set_params(self, *args: float, **kwargs: float) -> tuple[float]:
        """Set new parameters to the model.

        This works almost exactly as the unilateral model's
        :py:meth:`.Unilateral.set_params` method. However, this one allows the user to
        set the parameters of individual sides of the neck by prefixing the keyword
        arguments' names with ``"ipsi_"`` or ``"contra_"``.

        Anything not prefixed by ``"ipsi_"`` or ``"contra_"`` is passed to both sides
        of the neck. This does obviously not work with positional arguments.

        When setting the parameters via positional arguments, the order is
        important:

        1. The parameters of the edges from tumor to LNLs:

           1. first the ipsilateral parameters,
           2. if ``is_symmetric["tumor_spread"]`` is ``False``, the contralateral
              parameters. Otherwise, the ipsilateral parameters are used for both
              sides.

        2. The parameters of the edges from LNLs to tumor:

           1. again, first the ipsilateral parameters,
           2. if ``is_symmetric["lnl_spread"]`` is ``False``, the contralateral
              parameters. Otherwise, the ipsilateral parameters are used for both
              sides.

        3. The parameters of the parametric distributions for marginalizing over
           diagnose times.

        When still some positional arguments remain after that, they are returned
        in a tuple.
        """
        args = self.set_spread_params(*args, **kwargs)
        return self.set_distribution_params(*args, **kwargs)


    def load_patient_data(
        self,
        patient_data: pd.DataFrame,
        mapping: callable | dict[int, Any] = early_late_mapping,
    ) -> None:
        """Load patient data into the model.

        This amounts to calling the :py:meth:`~lymph.models.Unilateral.load_patient_data`
        method of both sides of the neck.
        """
        self.ipsi.load_patient_data(patient_data, "ipsi", mapping)
        self.contra.load_patient_data(patient_data, "contra", mapping)


    def comp_state_dist(
        self,
        t_stage: str = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> np.ndarray:
        """Compute the joint distribution over the ipsi- & contralateral hidden states.

        This computes the state distributions of both sides and returns their outer
        product. In case ``mode`` is ``"HMM"`` (default), the state distributions are
        first marginalized over the diagnose time distribtions of the respective
        ``t_stage``.

        See Also:
            :py:meth:`.Unilateral.comp_state_dist`
                The corresponding unilateral function. Note that this method returns
                a 2D array, because it computes the probability of any possible
                combination of ipsi- and contralateral states.
        """
        if mode == "HMM":
            ipsi_state_evo = self.ipsi.comp_dist_evolution()
            contra_state_evo = self.contra.comp_dist_evolution()
            time_marg_matrix = np.diag(self.get_distribution(t_stage).pmf)

            result = (
                ipsi_state_evo.T
                @ time_marg_matrix
                @ contra_state_evo
            )
        elif mode == "BN":
            ipsi_state_dist = self.ipsi.comp_state_dist(mode=mode)
            contra_state_dist = self.contra.comp_state_dist(mode=mode)

            result = np.outer(ipsi_state_dist, contra_state_dist)
        else:
            raise ValueError(f"Unknown mode '{mode}'.")

        return result


    def comp_obs_dist(
        self,
        t_stage: str = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> np.ndarray:
        """Compute the joint distribution over the ipsi- & contralateral observations.

        See Also:
            :py:meth:`.Unilateral.comp_obs_dist`
                The corresponding unilateral function. Note that this method returns
                a 2D array, because it computes the probability of any possible
                combination of ipsi- and contralateral observations.
        """
        joint_state_dist = self.comp_state_dist(t_stage=t_stage, mode=mode)
        return (
            self.ipsi.observation_matrix().T
            @ joint_state_dist
            @ self.contra.observation_matrix()
        )


    def comp_patient_llhs(
        self,
        t_stage: str = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> np.ndarray:
        """Compute the likelihood of each patient individually."""
        joint_state_dist = self.comp_state_dist(t_stage=t_stage, mode=mode)
        return matrix.fast_trace(
            self.ipsi.diagnose_matrices[t_stage].T,
            joint_state_dist @ self.contra.diagnose_matrices[t_stage],
        )


    def _bn_likelihood(self, log: bool = True, t_stage: str | None = None) -> float:
        """Compute the BN likelihood of data, using the stored params."""
        if t_stage is None:
            t_stage = "_BN"

        joint_state_dist = self.comp_state_dist(mode="BN")
        patient_llhs = np.sum(
            self.ipsi.diagnose_matrices[t_stage]
            * (joint_state_dist @ self.contra.diagnose_matrices[t_stage]),
            axis=0,
        )

        return np.sum(np.log(patient_llhs)) if log else np.prod(patient_llhs)


    def _hmm_likelihood(self, log: bool = True, t_stage: str | None = None) -> float:
        """Compute the HMM likelihood of data, using the stored params."""
        llh = 0. if log else 1.

        ipsi_dist_evo = self.ipsi.comp_dist_evolution()
        contra_dist_evo = self.contra.comp_dist_evolution()

        if t_stage is None:
            t_stages = self.t_stages
        else:
            t_stages = [t_stage]

        for stage in t_stages:
            diag_time_matrix = np.diag(self.get_distribution(stage).pmf)

            # Note that I am not using the `comp_joint_state_dist` method here, since
            # that would recompute the state dist evolution for each T-stage.
            joint_state_dist = (
                ipsi_dist_evo.T
                @ diag_time_matrix
                @ contra_dist_evo
            )
            patient_llhs = matrix.fast_trace(
                self.ipsi.diagnose_matrices[stage].T,
                joint_state_dist @ self.contra.diagnose_matrices[stage],
            )
            llh = add_or_mult(llh, patient_llhs, log)

        return llh


    def likelihood(
        self,
        given_params: Iterable[float] | dict[str, float] | None = None,
        log: bool = True,
        mode: Literal["HMM", "BN"] = "HMM",
        for_t_stage: str | None = None,
    ):
        """Compute the (log-)likelihood of the stored data given the model (and params).

        See the documentation of :py:meth:`.types.Model.likelihood` for more
        information on how to use the ``given_params`` parameter.

        Returns the log-likelihood if ``log`` is set to ``True``. The ``mode`` parameter
        determines whether the likelihood is computed for the hidden Markov model
        (``"HMM"``) or the Bayesian network (``"BN"``).

        Note:
            The computation is much faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.

        See Also:
            :py:meth:`.Unilateral.likelihood`
                The corresponding unilateral function.
        """
        try:
            # all functions and methods called here should raise a ValueError if the
            # given parameters are invalid...
            if given_params is None:
                pass
            elif isinstance(given_params, dict):
                self.set_params(**given_params)
            else:
                self.set_params(*given_params)
        except ValueError:
            return -np.inf if log else 0.

        if mode == "HMM":
            return self._hmm_likelihood(log, for_t_stage)
        if mode == "BN":
            return self._bn_likelihood(log, for_t_stage)

        raise ValueError("Invalid mode. Must be either 'HMM' or 'BN'.")


    def comp_posterior_joint_state_dist(
        self,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: dict[str, types.DiagnoseType] | None = None,
        t_stage: str | int = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> np.ndarray:
        """Compute joint post. dist. over ipsi & contra states, ``given_diagnoses``.

        The ``given_diagnoses`` is a dictionary storing a :py:class:`types.DiagnoseType`
        for the ``"ipsi"`` and ``"contra"`` side of the neck.

        Essentially, this is the risk for any possible combination of ipsi- and
        contralateral involvement, given the provided diagnoses.

        Note:
            The computation is much faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.

        See Also:
            :py:meth:`.Unilateral.comp_posterior_state_dist`
        """
        if isinstance(given_params, dict):
            self.set_params(**given_params)
        else:
            self.set_params(*given_params)

        if given_diagnoses is None:
            given_diagnoses = {}

        diagnose_given_state = {}
        for side in ["ipsi", "contra"]:
            if side not in given_diagnoses:
                warnings.warn(f"No diagnoses given for {side}lateral side.")

            diagnose_encoding = getattr(self, side).comp_diagnose_encoding(
                given_diagnoses.get(side, {})
            )
            observation_matrix = getattr(self, side).observation_matrix()
            # vector with P(Z=z|X) for each state X. A data matrix for one "patient"
            diagnose_given_state[side] = diagnose_encoding @ observation_matrix.T

        joint_state_dist = self.comp_state_dist(t_stage=t_stage, mode=mode)
        # matrix with P(Zi=zi,Zc=zc|Xi,Xc) * P(Xi,Xc) for all states Xi,Xc.
        joint_diagnose_and_state = np.outer(
            diagnose_given_state["ipsi"],
            diagnose_given_state["contra"],
        ) * joint_state_dist
        # Following Bayes' theorem, this is P(Xi,Xc|Zi=zi,Zc=zc) which is given by
        # P(Zi=zi,Zc=zc|Xi,Xc) * P(Xi,Xc) / P(Zi=zi,Zc=zc)
        return joint_diagnose_and_state / np.sum(joint_diagnose_and_state)


    def risk(
        self,
        involvement: types.PatternType | None = None,
        given_params: Iterable[float] | dict[str, float] | None = None,
        given_diagnoses: dict[str, types.DiagnoseType] | None = None,
        t_stage: str = "early",
        mode: Literal["HMM", "BN"] = "HMM",
    ) -> float:
        """Compute risk of an ``involvement`` pattern, given parameters and diagnoses.

        The parameters can be set via the ``given_params`` and ``given_params``, both
        of which are passed to the :py:meth:`.set_params` method. The
        ``given_diagnoses`` must be a dictionary mapping the side of the neck to a
        :py:class:`.types.DiagnoseType`.

        Note:
            The computation is much faster if no parameters are given, since then the
            transition matrix does not need to be recomputed.

        See Also:
            :py:meth:`.Unilateral.risk`
                The unilateral method for computing the risk of an involvment pattern.
            :py:meth:`.Bilateral.comp_posterior_joint_state_dist`
                This method computes the joint distribution over ipsi- and
                contralateral states, given the parameters and diagnoses. The risk then
                only marginalizes over the states that match the involvement pattern.
        """
        # TODO: test this method
        posterior_state_probs = self.comp_posterior_joint_state_dist(
            given_params=given_params,
            given_diagnoses=given_diagnoses,
            t_stage=t_stage,
            mode=mode,
        )

        if involvement is None:
            return posterior_state_probs

        marginalize_over_states = {}
        for side in ["ipsi", "contra"]:
            side_graph = getattr(self, side).graph
            marginalize_over_states[side] = matrix.compute_encoding(
                lnls=side_graph.lnls.keys(),
                pattern=involvement[side],
                base=3 if self.is_trinary else 2,
            )
        return (
            marginalize_over_states["ipsi"]
            @ posterior_state_probs
            @ marginalize_over_states["contra"]
        )


    def draw_patients(
        self,
        num: int,
        stage_dist: Iterable[float],
        rng: np.random.Generator | None = None,
        seed: int = 42,
        **_kwargs,
    ) -> pd.DataFrame:
        """Draw ``num`` random patients from the parametrized model.

        See Also:
            :py:meth:`.diagnose_times.Distribution.draw_diag_times`
                Method to draw diagnose times from a distribution.
            :py:meth:`.Unilateral.draw_diagnoses`
                Method to draw individual diagnoses from a unilateral model.
            :py:meth:`.Unilateral.draw_patients`
                The unilateral method to draw a synthetic dataset.
        """
        if rng is None:
            rng = np.random.default_rng(seed)

        if sum(stage_dist) != 1.:
            warnings.warn("Sum of stage distribution is not 1. Renormalizing.")
            stage_dist = np.array(stage_dist) / sum(stage_dist)

        drawn_t_stages = rng.choice(
            a=self.t_stages,
            p=stage_dist,
            size=num,
        )
        drawn_diag_times = [
            self.get_distribution(t_stage).draw_diag_times(rng=rng)
            for t_stage in drawn_t_stages
        ]

        drawn_obs_ipsi = self.ipsi.draw_diagnoses(drawn_diag_times, rng=rng)
        drawn_obs_contra = self.contra.draw_diagnoses(drawn_diag_times, rng=rng)
        drawn_obs = np.concatenate([drawn_obs_ipsi, drawn_obs_contra], axis=1)

        # construct MultiIndex with "ipsi" and "contra" at top level to allow
        # concatenation of the two separate drawn diagnoses
        sides = ["ipsi", "contra"]
        modality_names = list(self.get_all_modalities().keys())
        lnl_names = [lnl for lnl in self.ipsi.graph.lnls.keys()]
        multi_cols = pd.MultiIndex.from_product([sides, modality_names, lnl_names])

        # reorder the column levels and thus also the individual columns to match the
        # LyProX format without mixing up the data
        dataset = pd.DataFrame(drawn_obs, columns=multi_cols)
        dataset = dataset.reorder_levels(order=[1, 0, 2], axis="columns")
        dataset = dataset.sort_index(axis="columns", level=0)
        dataset[('tumor', '1', 't_stage')] = drawn_t_stages

        return dataset



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/main/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from pkg_resources import DistributionNotFound, get_distribution

sys.path.insert(0, os.path.abspath('../..'))

try:
    __version__ = get_distribution("lymph").version
except DistributionNotFound:
    __version__ = "unknown version"


# -- Project information -----------------------------------------------------

project = 'lymph'
copyright = '2022, Roman Ludwig'
author = 'Roman Ludwig'
gh_username = 'rmnldwg'

version = __version__
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.intersphinx',
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'myst_nb'
]

# MyST settings
myst_enable_extensions = ["colon_fence", "dollarmath"]
nb_execution_mode = "auto"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_data']

# document classes and their constructors
autoclass_content = 'class'

# sort members by source
autodoc_member_order = 'bysource'

# show type hints
autodoc_typehints = 'signature'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = 'sphinx_book_theme'
html_theme_options = {
    "repository_url": f"https://github.com/{gh_username}/{project}",
    "repository_branch": "main",
    "use_repository_button": True,
}

# import sphinx_modern_theme
# html_theme = "sphinx_modern_theme"
# html_theme_path = [sphinx_modern_theme.get_html_theme_path()]

# html_theme = "bootstrap-astropy"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['./_static']
html_css_files = [
    "css/custom.css",
]



"""
Test helper functions mainly by executing their doctests.
"""
import doctest

from lymph import helper


def load_tests(loader, tests, ignore):
    """Add doctests to the unittests."""
    tests.addTests(doctest.DocTestSuite(helper))
    return tests
